import os
from typing import Optional
from unittest.mock import ANY, Mock, call, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import JobState, JobStatus
from oumi.launcher.clients.sky_client import (
    SkyClient,
    _convert_job_to_task,
    _get_use_spot_vm_override,
)


#
# Fixtures
#
@pytest.fixture
def mock_sky_data_storage():
    with patch("sky.data.Storage") as mock_storage:
        yield mock_storage


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80GB",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
        image_id="docker://ubuntu:latest",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={},
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup="pip install -r requirements.txt",
        run="./hello_world.sh",
    )


#
# Tests
#
def test_sky_client_gcp_name():
    client = SkyClient()
    assert client.SupportedClouds.GCP.value == "gcp"


def test_sky_client_runpod_name():
    client = SkyClient()
    assert client.SupportedClouds.RUNPOD.value == "runpod"


def test_sky_client_lambda_name():
    client = SkyClient()
    assert client.SupportedClouds.LAMBDA.value == "lambda"


def test_sky_client_aws_name():
    client = SkyClient()
    assert client.SupportedClouds.AWS.value == "aws"


def test_sky_client_azure_name():
    client = SkyClient()
    assert client.SupportedClouds.AZURE.value == "azure"


def test_convert_job_to_task(
    mock_sky_data_storage,
):
    with patch.dict(os.environ, {"OUMI_USE_SPOT_VM": "nonspot"}, clear=True):
        with patch("sky.Resources") as mock_resources:
            with patch("sky.clouds.GCP") as mock_cloud:
                mock_gcp = Mock()
                mock_cloud.return_value = mock_gcp
                with patch("sky.Task") as mock_task_cls:
                    mock_task = Mock()
                    mock_task_cls.return_value = mock_task
                    job = _get_default_job("gcp")
                    _ = _convert_job_to_task(job)
                    mock_resources.assert_has_calls(
                        [
                            call(
                                cloud=mock_gcp,
                                instance_type=job.resources.instance_type,
                                cpus=job.resources.cpus,
                                memory=job.resources.memory,
                                accelerators=job.resources.accelerators,
                                use_spot=False,
                                region=job.resources.region,
                                zone=job.resources.zone,
                                disk_size=job.resources.disk_size,
                                disk_tier=job.resources.disk_tier,
                                image_id=job.resources.image_id,
                            )
                        ]
                    )
                    mock_task_cls.assert_has_calls(
                        [
                            call(
                                name=job.name,
                                setup=job.setup,
                                run=job.run,
                                envs=job.envs,
                                workdir=job.working_dir,
                                num_nodes=job.num_nodes,
                            )
                        ]
                    )
                    mock_task.set_file_mounts.assert_called_once()
                    mock_task.set_storage_mounts.assert_called_once()
                    mock_task.set_resources.assert_called_once()


def test_convert_job_to_task_with_dict_image_id(
    mock_sky_data_storage,
):
    with patch.dict(os.environ, {"OUMI_USE_SPOT_VM": "nonspot"}, clear=True):
        with patch("sky.Resources") as mock_resources:
            with patch("sky.clouds.GCP") as mock_cloud:
                mock_gcp = Mock()
                mock_cloud.return_value = mock_gcp
                with patch("sky.Task") as mock_task_cls:
                    mock_task = Mock()
                    mock_task_cls.return_value = mock_task
                    job = _get_default_job("gcp")

                    # Test with image_id_map and empty image_id.
                    # Should use image_id_map.
                    job.resources.image_id = None
                    job.resources.image_id_map = {
                        "us-central1": "ami-0c1a2b3c4d5e6f701",
                        "us-east1": "ami-0d2e3f4g5h6i7j802",
                        "us-west1": "ami-0e3f4g5h6i7j8k903",
                        "us-west2": "ami-0f4g5h6i7j8k9l004",
                        "us-west3": "ami-0a5b6c7d8e9f0g105",
                        "us-west4": "ami-0b6c7d8e9f0g1h206",
                        "us-west5": "ami-0c7d8e9f0g1h2i307",
                        "us-west6": "ami-0d8e9f0g1h2i3j408",
                    }

                    _ = _convert_job_to_task(job)
                    mock_resources.assert_has_calls(
                        [
                            call(
                                cloud=mock_gcp,
                                instance_type=job.resources.instance_type,
                                cpus=job.resources.cpus,
                                memory=job.resources.memory,
                                accelerators=job.resources.accelerators,
                                use_spot=False,
                                region=job.resources.region,
                                zone=job.resources.zone,
                                disk_size=job.resources.disk_size,
                                disk_tier=job.resources.disk_tier,
                                image_id=job.resources.image_id_map,
                            )
                        ]
                    )
                    mock_task_cls.assert_has_calls(
                        [
                            call(
                                name=job.name,
                                setup=job.setup,
                                run=job.run,
                                envs=job.envs,
                                workdir=job.working_dir,
                                num_nodes=job.num_nodes,
                            )
                        ]
                    )
                    mock_task.set_file_mounts.assert_called_once()
                    mock_task.set_storage_mounts.assert_called_once()
                    mock_task.set_resources.assert_called_once()


def test_convert_job_to_task_with_populated_image_and_dict(
    mock_sky_data_storage,
):
    with patch.dict(os.environ, {"OUMI_USE_SPOT_VM": "nonspot"}, clear=True):
        with patch("sky.Resources") as mock_resources:
            with patch("sky.clouds.GCP") as mock_cloud:
                mock_gcp = Mock()
                mock_cloud.return_value = mock_gcp
                with patch("sky.Task") as mock_task_cls:
                    mock_task = Mock()
                    mock_task_cls.return_value = mock_task
                    job = _get_default_job("gcp")

                    # Test with populated image_id_map and image_id.
                    # Should use image_id.
                    job.resources.image_id_map = {
                        "us-central1": "ami-0c1a2b3c4d5e6f701",
                        "us-east1": "ami-0d2e3f4g5h6i7j802",
                        "us-west1": "ami-0e3f4g5h6i7j8k903",
                        "us-west2": "ami-0f4g5h6i7j8k9l004",
                        "us-west3": "ami-0a5b6c7d8e9f0g105",
                        "us-west4": "ami-0b6c7d8e9f0g1h206",
                        "us-west5": "ami-0c7d8e9f0g1h2i307",
                        "us-west6": "ami-0d8e9f0g1h2i3j408",
                    }
                    _ = _convert_job_to_task(job)
                    mock_resources.assert_has_calls(
                        [
                            call(
                                cloud=mock_gcp,
                                instance_type=job.resources.instance_type,
                                cpus=job.resources.cpus,
                                memory=job.resources.memory,
                                accelerators=job.resources.accelerators,
                                use_spot=False,
                                region=job.resources.region,
                                zone=job.resources.zone,
                                disk_size=job.resources.disk_size,
                                disk_tier=job.resources.disk_tier,
                                image_id=job.resources.image_id,
                            )
                        ]
                    )
                    mock_task_cls.assert_has_calls(
                        [
                            call(
                                name=job.name,
                                setup=job.setup,
                                run=job.run,
                                envs=job.envs,
                                workdir=job.working_dir,
                                num_nodes=job.num_nodes,
                            )
                        ]
                    )
                    mock_task.set_file_mounts.assert_called_once()
                    mock_task.set_storage_mounts.assert_called_once()
                    mock_task.set_resources.assert_called_once()


@pytest.mark.parametrize(
    "env_var_use_spot_vm,expected_use_spot_vm",
    [
        (None, None),
        ("spot", True),
        ("preemptable", True),
        ("nonspot", False),
        ("non-preemptible", False),
    ],
)
def test_get_use_spot_vm_override(
    env_var_use_spot_vm: Optional[str], expected_use_spot_vm: bool
):
    if env_var_use_spot_vm is not None:
        with patch.dict(
            os.environ, {"OUMI_USE_SPOT_VM": env_var_use_spot_vm}, clear=True
        ):
            assert _get_use_spot_vm_override() == expected_use_spot_vm
    else:
        with patch.dict(os.environ, {}, clear=True):
            assert _get_use_spot_vm_override() is None


def test_sky_client_launch(
    mock_sky_data_storage,
):
    with patch("sky.launch") as mock_launch:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            job = _get_default_job("gcp")
            mock_resource_handle = Mock()
            mock_resource_handle.cluster_name = "mycluster"
            mock_launch.return_value = (1, mock_resource_handle)
            mock_stream_and_get.return_value = (1, mock_resource_handle)
            client = SkyClient()
            job_status = client.launch(job)
            expected_job_status = JobStatus(
                name="",
                id="1",
                cluster="mycluster",
                status="",
                metadata="",
                done=False,
                state=JobState.PENDING,
            )
            assert job_status == expected_job_status
            mock_launch.assert_called_once_with(
                ANY,
                cluster_name=None,
                idle_minutes_to_autostop=60,
            )


def test_sky_client_launch_no_stop(
    mock_sky_data_storage,
):
    with patch("sky.launch") as mock_launch:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            job = _get_default_job("runpod")
            job.resources.region = "ca"
            job.resources.disk_tier = "best"
            mock_resource_handle = Mock()
            mock_resource_handle.cluster_name = "mycluster"
            mock_launch.return_value = (1, mock_resource_handle)
            mock_stream_and_get.return_value = (1, mock_resource_handle)
            client = SkyClient()
            job_status = client.launch(job)
            expected_job_status = JobStatus(
                name="",
                id="1",
                cluster="mycluster",
                status="",
                metadata="",
                done=False,
                state=JobState.PENDING,
            )
            assert job_status == expected_job_status
            mock_launch.assert_called_once_with(
                ANY,
                cluster_name=None,
                idle_minutes_to_autostop=None,
            )


def test_sky_client_launch_kwarg(mock_sky_data_storage):
    with patch("sky.launch") as mock_launch:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            job = _get_default_job("gcp")
            mock_resource_handle = Mock()
            mock_resource_handle.cluster_name = "mycluster"
            mock_launch.return_value = (1, mock_resource_handle)
            mock_stream_and_get.return_value = (1, mock_resource_handle)
            client = SkyClient()
            job_status = client.launch(job, idle_minutes_to_autostop=None)
            expected_job_status = JobStatus(
                name="",
                id="1",
                cluster="mycluster",
                status="",
                metadata="",
                done=False,
                state=JobState.PENDING,
            )
            assert job_status == expected_job_status
            mock_launch.assert_called_once_with(
                ANY,
                cluster_name=None,
                idle_minutes_to_autostop=None,
            )


def test_sky_client_launch_kwarg_value(mock_sky_data_storage):
    with patch("sky.launch") as mock_launch:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            job = _get_default_job("gcp")
            mock_resource_handle = Mock()
            mock_resource_handle.cluster_name = "mycluster"
            mock_launch.return_value = (1, mock_resource_handle)
            mock_stream_and_get.return_value = (1, mock_resource_handle)
            client = SkyClient()
            job_status = client.launch(job, idle_minutes_to_autostop=45)
            expected_job_status = JobStatus(
                name="",
                id="1",
                cluster="mycluster",
                status="",
                metadata="",
                done=False,
                state=JobState.PENDING,
            )
            assert job_status == expected_job_status
            mock_launch.assert_called_once_with(
                ANY,
                cluster_name=None,
                idle_minutes_to_autostop=45,
            )


def test_sky_client_launch_unused_kwarg(mock_sky_data_storage):
    with patch("sky.launch") as mock_launch:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            job = _get_default_job("gcp")
            mock_resource_handle = Mock()
            mock_resource_handle.cluster_name = "mycluster"
            mock_launch.return_value = (1, mock_resource_handle)
            mock_stream_and_get.return_value = (1, mock_resource_handle)
            client = SkyClient()
            job_status = client.launch(job, foo=None)
            expected_job_status = JobStatus(
                name="",
                id="1",
                cluster="mycluster",
                status="",
                metadata="",
                done=False,
                state=JobState.PENDING,
            )
            assert job_status == expected_job_status
            mock_launch.assert_called_once_with(
                ANY,
                cluster_name=None,
                idle_minutes_to_autostop=60,
            )


def test_sky_client_launch_with_cluster_name(mock_sky_data_storage):
    with patch("sky.launch") as mock_launch:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            job = _get_default_job("gcp")
            mock_resource_handle = Mock()
            mock_resource_handle.cluster_name = "cluster_name"
            mock_launch.return_value = (1, mock_resource_handle)
            mock_stream_and_get.return_value = (1, mock_resource_handle)
            client = SkyClient()
            job_status = client.launch(job, "cluster_name")
            expected_job_status = JobStatus(
                name="",
                id="1",
                cluster="cluster_name",
                status="",
                metadata="",
                done=False,
                state=JobState.PENDING,
            )
            assert job_status == expected_job_status
            mock_launch.assert_called_once_with(
                ANY,
                cluster_name="cluster_name",
                idle_minutes_to_autostop=60,
            )


def test_sky_client_status():
    with patch("sky.status") as mock_status:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            mock_status.return_value = [{"name": "mycluster"}]
            mock_stream_and_get.return_value = [{"name": "mycluster"}]
            client = SkyClient()
            status = client.status()
            mock_status.assert_called_once()
            mock_stream_and_get.assert_called_once_with([{"name": "mycluster"}])
            assert status == [{"name": "mycluster"}]


def test_sky_client_queue():
    with patch("sky.queue") as mock_queue:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            mock_queue.return_value = [{"name": "myjob"}]
            mock_stream_and_get.return_value = [{"name": "myjob"}]
            client = SkyClient()
            queue = client.queue("mycluster")
            mock_queue.assert_called_once_with("mycluster")
            mock_stream_and_get.assert_called_once_with([{"name": "myjob"}])
            assert queue == [{"name": "myjob"}]


def test_sky_client_cancel():
    with patch("sky.cancel") as mock_cancel:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            client = SkyClient()
            client.cancel("mycluster", "1")
            mock_cancel.assert_called_once_with(cluster_name="mycluster", job_ids=[1])
            mock_stream_and_get.assert_called_once()


def test_sky_client_exec(mock_sky_data_storage):
    with patch("sky.exec") as mock_exec:
        with patch("sky.stream_and_get") as mock_stream_and_get:
            mock_resource_handle = Mock()
            mock_exec.return_value = (1, mock_resource_handle)
            mock_stream_and_get.return_value = (1, mock_resource_handle)
            client = SkyClient()
            job = _get_default_job("gcp")
            job_id = client.exec(job, "mycluster")
            mock_exec.assert_called_once_with(ANY, "mycluster")
            mock_stream_and_get.assert_called_once_with((1, mock_resource_handle))
            assert job_id == "1"


def test_sky_client_exec_runtime_error(mock_sky_data_storage):
    with pytest.raises(RuntimeError):
        with patch("sky.exec") as mock_exec:
            with patch("sky.stream_and_get") as mock_stream_and_get:
                mock_resource_handle = Mock()
                mock_exec.return_value = (None, mock_resource_handle)
                mock_stream_and_get.return_value = (None, mock_resource_handle)
                client = SkyClient()
                job = _get_default_job("gcp")
                _ = client.exec(job, "mycluster")


def test_sky_client_down():
    with patch("sky.down") as mock_down:
        client = SkyClient()
        client.down("mycluster")
        mock_down.assert_called_once_with("mycluster")


def test_sky_client_stop():
    with patch("sky.stop") as mock_stop:
        client = SkyClient()
        client.stop("mycluster")
        mock_stop.assert_called_once_with("mycluster")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch

import jsonlines
import pandas as pd
import pytest

from oumi.core.configs.analyze_config import (
    AnalyzeConfig,
    DatasetSource,
)
from oumi.core.datasets import BaseMapDataset
from oumi.datasets import TextSftJsonLinesDataset, VLJsonlinesDataset
from oumi.utils.analysis_utils import (
    build_tokenizer_from_config,
    compute_statistics,
    load_dataset_from_config,
)


@pytest.fixture
def mock_dataset_class_and_instance():
    """Fixture to create mock dataset class and instance."""
    mock_dataset_class = Mock()
    mock_dataset_instance = Mock(spec=BaseMapDataset)
    mock_dataset_class.return_value = mock_dataset_instance
    return mock_dataset_class, mock_dataset_instance


@pytest.fixture
def mock_registry():
    """Fixture to patch the registry."""
    with patch("oumi.utils.analysis_utils.REGISTRY") as mock_registry:
        yield mock_registry


@pytest.fixture
def sample_conversation_data():
    """Sample conversation format data."""
    return [
        {
            "messages": [
                {"role": "user", "content": "Hello, how are you?"},
                {"role": "assistant", "content": "I'm doing well, thank you!"},
            ]
        },
        {
            "messages": [
                {"role": "user", "content": "What's the weather like?"},
                {
                    "role": "assistant",
                    "content": "I don't have access to real-time weather data.",
                },
            ]
        },
    ]


@pytest.fixture
def sample_alpaca_data():
    """Sample alpaca format data."""
    return [
        {
            "instruction": "What's the weather like in Seattle today?",
            "input": "",
            "output": "I apologize, but I don't have access to real-time weather "
            "information for Seattle.",
        },
        {
            "instruction": "Compute the average of the presented numbers.",
            "input": "5, 6, 10",
            "output": "The average for the numbers: 5, 6, 10 can be computed by "
            "adding first all of them, and then dividing this sum by their total "
            "number. First, 5+6+10 = 21. Then, 21 / 3 = 7. The average is 7.",
        },
    ]


@pytest.fixture
def sample_vision_language_data():
    """Sample vision-language data."""
    return [
        {
            "messages": [
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "content": "https://example.com/image_of_dog.jpg",
                        },
                        {"type": "text", "content": "What breed is this dog?"},
                    ],
                },
                {
                    "role": "assistant",
                    "content": "This appears to be a Shih Tzu puppy.",
                },
            ]
        },
        {
            "messages": [
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "content": "https://example.com/scenic_view.jpg",
                        },
                        {"type": "text", "content": "Describe this image:"},
                    ],
                },
                {
                    "role": "assistant",
                    "content": "A scenic view of the puget sound with mountains in "
                    "the background.",
                },
            ]
        },
    ]


@pytest.fixture
def temp_conversation_file(sample_conversation_data):
    """Create a temporary conversation format file."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        json.dump(sample_conversation_data, f)
        f.flush()  # Ensure data is written to disk
        yield f.name
    Path(f.name).unlink()  # Cleanup


@pytest.fixture
def temp_alpaca_file(sample_alpaca_data):
    """Create a temporary alpaca format file."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        json.dump(sample_alpaca_data, f)
        f.flush()  # Ensure data is written to disk
        yield f.name
    Path(f.name).unlink()  # Cleanup


@pytest.fixture
def temp_vision_language_file(sample_vision_language_data):
    """Create a temporary vision-language format file."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".jsonl", delete=False) as f:
        with jsonlines.open(f.name, mode="w") as writer:
            writer.write_all(sample_vision_language_data)
        yield f.name
    Path(f.name).unlink()  # Cleanup


def test_load_dataset_from_config_success(
    mock_dataset_class_and_instance, mock_registry
):
    """Test successful dataset loading."""
    config = AnalyzeConfig(
        dataset_name="test_dataset",
        split="train",
    )

    mock_dataset_class, mock_dataset_instance = mock_dataset_class_and_instance
    mock_registry.get_dataset.return_value = mock_dataset_class

    result = load_dataset_from_config(config)

    assert result == mock_dataset_instance
    assert mock_registry.get_dataset.called


def test_load_dataset_from_config_missing_dataset_name():
    """Test error handling when dataset_name is not provided."""
    with pytest.raises(
        ValueError,
        match="Either 'dataset_name' or 'dataset_path' must be provided when "
        "dataset_source=DatasetSource.CONFIG",
    ):
        AnalyzeConfig(
            dataset_source=DatasetSource.CONFIG,  # Required field
            dataset_name=None,
            dataset_path=None,
            split="train",
        )


def test_load_dataset_from_config_dataset_not_registered(mock_registry):
    """Test error handling when dataset is not found in registry."""
    config = AnalyzeConfig(
        dataset_name="nonexistent_dataset",
        split="train",
    )

    mock_registry.get_dataset.return_value = None

    with pytest.raises(
        NotImplementedError,
        match=(
            "Dataset 'nonexistent_dataset' is not registered in the REGISTRY. "
            "Loading from HuggingFace Hub is not yet implemented."
        ),
    ):
        load_dataset_from_config(config)


def test_load_dataset_from_config_for_non_basemapdataset(mock_registry):
    """Test error handling when dataset class doesn't inherit from BaseMapDataset."""
    config = AnalyzeConfig(
        dataset_name="test_dataset",
        split="train",
    )

    mock_dataset_class = Mock()
    mock_dataset_instance = Mock()  # Not a BaseMapDataset
    mock_dataset_class.return_value = mock_dataset_instance

    mock_registry.get_dataset.return_value = mock_dataset_class

    with pytest.raises(
        NotImplementedError,
        match=(
            "Dataset type .* is not supported for analysis. "
            "Please use a dataset that inherits from BaseMapDataset."
        ),
    ):
        load_dataset_from_config(config)


def test_load_dataset_from_config_registry_exception(mock_registry):
    """Test error handling when registry.get_dataset raises an exception."""
    config = AnalyzeConfig(
        dataset_name="test_dataset",
        split="train",
    )

    mock_registry.get_dataset.side_effect = Exception("Registry error")

    with pytest.raises(Exception, match="Registry error"):
        load_dataset_from_config(config)


def test_load_dataset_from_config_with_processor_parameters(
    mock_dataset_class_and_instance, mock_registry
):
    """Test dataset loading with processor parameters."""
    config = AnalyzeConfig(
        dataset_name="test_dataset",
        processor_name="Salesforce/blip2-opt-2.7b",
        processor_kwargs={"image_size": 224},
        trust_remote_code=True,
    )

    mock_dataset_class, mock_dataset_instance = mock_dataset_class_and_instance
    mock_registry.get_dataset.return_value = mock_dataset_class

    result = load_dataset_from_config(config)

    # Verify the dataset was called with processor parameters
    mock_dataset_class.assert_called_once()
    call_kwargs = mock_dataset_class.call_args[1]
    assert call_kwargs["processor_name"] == "Salesforce/blip2-opt-2.7b"
    assert call_kwargs["processor_kwargs"] == {"image_size": 224}
    assert call_kwargs["trust_remote_code"] is True
    assert result == mock_dataset_instance


def test_build_tokenizer_from_config_success():
    """Test successful tokenizer building from config."""
    tokenizer_config = {
        "model_name": "gpt2",
        "tokenizer_kwargs": {"padding_side": "left"},
        "trust_remote_code": False,
    }

    tokenizer = build_tokenizer_from_config(tokenizer_config)

    assert tokenizer is not None
    assert hasattr(tokenizer, "encode")
    assert hasattr(tokenizer, "decode")


def test_build_tokenizer_from_config_none():
    """Test tokenizer building with None config."""
    tokenizer = build_tokenizer_from_config(None)

    assert tokenizer is None


def test_build_tokenizer_from_config_missing_model_name():
    """Test error handling when model_name is missing from config."""
    tokenizer_config = {
        "tokenizer_kwargs": {"padding_side": "left"},
        "trust_remote_code": False,
    }

    with pytest.raises(
        ValueError, match="tokenizer_config must contain 'model_name' field"
    ):
        build_tokenizer_from_config(tokenizer_config)


def test_load_dataset_from_config_with_tokenizer(
    mock_dataset_class_and_instance, mock_registry
):
    """Test dataset loading with provided tokenizer."""
    config = AnalyzeConfig(
        dataset_name="test_dataset",
        split="train",
    )

    mock_tokenizer = Mock()
    mock_dataset_class, mock_dataset_instance = mock_dataset_class_and_instance
    mock_registry.get_dataset.return_value = mock_dataset_class

    result = load_dataset_from_config(config, tokenizer=mock_tokenizer)

    # Verify the dataset was called with tokenizer
    mock_dataset_class.assert_called_once()
    call_kwargs = mock_dataset_class.call_args[1]
    assert call_kwargs["tokenizer"] == mock_tokenizer
    assert result == mock_dataset_instance


def test_load_dataset_from_config_without_tokenizer(
    mock_dataset_class_and_instance, mock_registry
):
    """Test dataset loading without tokenizer."""
    config = AnalyzeConfig(
        dataset_name="test_dataset",
        split="train",
    )

    mock_dataset_class, mock_dataset_instance = mock_dataset_class_and_instance
    mock_registry.get_dataset.return_value = mock_dataset_class

    result = load_dataset_from_config(config)

    # Verify the dataset was called without tokenizer
    mock_dataset_class.assert_called_once()
    call_kwargs = mock_dataset_class.call_args[1]
    assert "tokenizer" not in call_kwargs
    assert result == mock_dataset_instance


# Custom dataset loading tests
def test_load_custom_dataset_conversation_format(temp_conversation_file):
    """Test loading custom dataset in conversation format."""
    config = AnalyzeConfig(
        dataset_path=temp_conversation_file,
        dataset_format="oumi",
        is_multimodal=False,  # Explicitly set as text-only
    )

    dataset = load_dataset_from_config(config)

    assert isinstance(dataset, TextSftJsonLinesDataset)
    assert len(dataset) == 2

    # Check first conversation
    conv1 = dataset.conversation(0)
    assert len(conv1.messages) == 2
    assert conv1.messages[0].role == "user"
    assert conv1.messages[0].content == "Hello, how are you?"
    assert conv1.messages[1].role == "assistant"
    assert conv1.messages[1].content == "I'm doing well, thank you!"


def test_load_custom_dataset_alpaca_format(temp_alpaca_file):
    """Test loading custom dataset in alpaca format."""
    config = AnalyzeConfig(
        dataset_path=temp_alpaca_file,
        dataset_format="alpaca",
        is_multimodal=False,  # Explicitly set as text-only
    )

    dataset = load_dataset_from_config(config)

    assert isinstance(dataset, TextSftJsonLinesDataset)
    assert len(dataset) == 2

    # Check conversation structure
    conv1 = dataset.conversation(0)
    assert len(conv1.messages) == 2
    assert conv1.messages[0].role == "user"
    assert conv1.messages[1].role == "assistant"


def test_load_custom_dataset_multi_modal(temp_vision_language_file):
    """Test loading custom multimodal dataset with processor."""
    # Create a mock tokenizer with required attributes
    mock_tokenizer = Mock()
    mock_tokenizer.pad_token_id = 0  # Set a valid pad_token_id

    config = AnalyzeConfig(
        dataset_path=temp_vision_language_file,
        dataset_format="oumi",
        processor_name="openai/clip-vit-base-patch32",  # Processor provided
        is_multimodal=True,  # Explicitly mark as multimodal
    )

    dataset = load_dataset_from_config(config, tokenizer=mock_tokenizer)
    assert isinstance(dataset, VLJsonlinesDataset)
    assert len(dataset) == 2


def test_load_custom_dataset_text(temp_conversation_file):
    """Test that text-only datasets are correctly detected and loaded as
    TextSftJsonLinesDataset."""
    config = AnalyzeConfig(
        dataset_path=temp_conversation_file,
        dataset_format="oumi",
        is_multimodal=False,  # Explicitly set as text-only
    )

    dataset = load_dataset_from_config(config)

    # Should be detected as text-only and loaded as TextSftJsonLinesDataset
    assert isinstance(dataset, TextSftJsonLinesDataset)
    assert len(dataset) == 2


def test_load_custom_dataset_file_not_found():
    """Test error handling when custom dataset file doesn't exist."""
    config = AnalyzeConfig(
        dataset_path="nonexistent_file.json",
        dataset_format="oumi",  # Required for custom datasets
        is_multimodal=False,  # Required for custom datasets
    )

    with pytest.raises(
        FileNotFoundError, match="Dataset file not found: nonexistent_file.json"
    ):
        load_dataset_from_config(config)


def test_load_custom_dataset_directory_path():
    """Test error handling when custom dataset path is a directory."""
    with tempfile.TemporaryDirectory() as temp_dir:
        config = AnalyzeConfig(
            dataset_path=temp_dir,
            dataset_format="oumi",  # Required for custom datasets
            is_multimodal=False,  # Required for custom datasets
        )

        with pytest.raises(
            ValueError, match="Dataset path must be a file, not a directory"
        ):
            load_dataset_from_config(config)


def test_compute_statistics_empty_series():
    """Test compute_statistics with empty pandas Series."""
    empty_series = pd.Series([], dtype=float)
    result = compute_statistics(empty_series)

    expected = {
        "count": 0,
        "mean": 0.0,
        "std": 0.0,
        "min": 0,
        "max": 0,
        "median": 0.0,
    }
    assert result == expected


def test_compute_statistics_single_value():
    """Test compute_statistics with single value (edge case for NaN std)."""
    single_series = pd.Series([42.5])
    result = compute_statistics(single_series)

    expected = {
        "count": 1,
        "mean": 42.5,
        "std": 0.0,  # Standard deviation is 0 for single value
        "min": 42.5,
        "max": 42.5,
        "median": 42.5,
    }
    assert result == expected


def test_compute_statistics_multiple_values():
    """Test compute_statistics with multiple values."""
    series = pd.Series([1, 2, 3, 4, 5])
    result = compute_statistics(series)

    expected = {
        "count": 5,
        "mean": 3.0,
        "std": 1.58,
        "min": 1,
        "max": 5,
        "median": 3.0,
    }
    assert result == expected


def test_compute_statistics_multiple_values_with_precision():
    """Test compute_statistics with multiple values and custom decimal precision."""
    series = pd.Series([1.1, 2.2, 3.3, 4.4, 5.5])
    result = compute_statistics(series, decimal_precision=1)

    expected = {
        "count": 5,
        "mean": 3.3,
        "std": 1.7,
        "min": 1.1,
        "max": 5.5,
        "median": 3.3,
    }
    assert result == expected



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional, Union

import pandas as pd
from typing_extensions import override

from oumi.core.datasets import VisionLanguageSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation
from oumi.utils.io_utils import load_jsonlines


@register_dataset("vl_sft")
@register_dataset("vision_language_jsonl")
class VLJsonlinesDataset(VisionLanguageSftDataset):
    """VLJsonlinesDataset for loading Vision-Language SFT data in Oumi format.

    This dataset class is designed to work with JSON Lines (.jsonl) files containing
    Vision-Language supervised fine-tuning (SFT) data. It supports loading data either
    from a file or from a provided list of data samples.

    Usage example:
        Examples:
            Loading from a file:
                >>> from oumi.datasets import VLJsonlinesDataset
                >>> dataset = VLJsonlinesDataset( # doctest: +SKIP
                ...     dataset_path="/path/to/your/dataset.jsonl",
                ... )

            Loading from a list of data samples:
                >>> from oumi.builders import build_processor, build_tokenizer
                >>> from oumi.core.configs import ModelParams
                >>> from oumi.datasets import VLJsonlinesDataset
                >>> data_samples = [
                ...     {
                ...         "messages": [
                ...             {
                ...                 "role": "user",
                ...                 "content": "Describe this image:",
                ...                 "type": "text"
                ...             },
                ...             {
                ...                 "role": "user",
                ...                 "content": "path/to/image.jpg",
                ...                 "type": "image_path"
                ...             },
                ...             {
                ...                 "role": "assistant",
                ...                 "content": "A scenic view of the puget sound.",
                ...                 "type": "text",
                ...             },
                ...         ]
                ...     }
                ... ]
                >>> tokenizer = build_tokenizer(
                ...     ModelParams(model_name="Qwen/Qwen2-1.5B-Instruct")
                ... )
                >>> dataset = VLJsonlinesDataset(
                ...     data=data_samples,
                ...     tokenizer=tokenizer,
                ...     processor_name="openai/clip-vit-base-patch32",
                ... )
    """

    default_dataset = "custom"

    def __init__(
        self,
        dataset_path: Optional[Union[str, Path]] = None,
        data: Optional[list] = None,
        **kwargs,
    ):
        """Initializes a new instance of the VLJsonlinesDataset class."""
        if dataset_path is not None and data is not None:
            raise ValueError(
                "Either dataset_path or data must be provided, but not both"
            )

        self._data_column: str = "_messages_column"
        self._dataset_path: Optional[Path] = (
            Path(dataset_path) if dataset_path else None
        )

        if data is not None:
            data_frame = pd.DataFrame({self._data_column: data})
        elif self._dataset_path is not None:
            data = load_jsonlines(self._dataset_path)
            data_frame = pd.DataFrame({self._data_column: data})
        else:
            raise ValueError("Dataset path or data must be provided")

        assert data_frame is not None
        self._data: pd.DataFrame = data_frame

        super().__init__(**kwargs)

    @override
    def _load_data(self) -> pd.DataFrame:
        # no-op, data is already loaded in __init__
        return self._data

    @override
    def transform_conversation(self, example: dict) -> Conversation:
        """Transform a single conversation example into a Conversation object."""
        messages = example[self._data_column]
        return Conversation.from_dict(messages)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("allenai/dolma")
class DolmaDataset(BasePretrainingDataset):
    """Dolma: A dataset of 3 trillion tokens from diverse web content.

    Dolma :footcite:`2024_dolma` is a large-scale dataset containing
    approximately 3 trillion tokens sourced from various web content, academic
    publications, code, books, and encyclopedic materials. It is designed for
    language modeling tasks and casual language model training.

    The dataset is available in multiple versions, with v1.7 being the latest
    release used to train OLMo 7B-v1.7. It includes data from sources such as
    Common Crawl, Refined Web, StarCoder, C4, Reddit, Semantic Scholar, arXiv,
    StackExchange, and more.

    Data Fields:
      id (str): Unique identifier for the data entry.
      text (str): The main content of the data entry.
      added (str, optional): Timestamp indicating when the entry was added
        to the dataset.
      created (str, optional): Timestamp indicating when the original content
        was created.
      source (str, optional): Information about the origin or source of the
        data.

    See Also:
      - Paper: https://arxiv.org/abs/2402.00159
      - GitHub project: https://github.com/allenai/dolma
      - Hugging Face Hub: https://huggingface.co/datasets/allenai/dolma

    Note:
        The dataset is released under the ODC-BY license. Users are bound by
        the license agreements and terms of use of the original data sources.

    Citations:
      .. footbibliography::
    """

    default_dataset = "allenai/dolma"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Volcano Engine Reinforcement Learning (verl) GRPO Trainer."""

import copy
import os
from pathlib import Path
from pprint import pformat
from typing import Any, Callable, Optional, Union, cast

from datasets import Dataset
from omegaconf import DictConfig, OmegaConf

from oumi.core.types.conversation import Conversation
from oumi.utils.grpo_utils import (
    extract_prompt_images_completion_from_single_turn_conversation,
)

try:
    import ray  # pyright: ignore[reportMissingImports]
    import verl  # pyright: ignore[reportMissingImports]
    from verl.trainer.ppo.ray_trainer import (  # pyright: ignore[reportMissingImports]
        RayPPOTrainer,
        ResourcePoolManager,
        Role,
    )
    from verl.workers.fsdp_workers import (  # pyright: ignore[reportMissingImports]
        ActorRolloutRefWorker,
        CriticWorker,
    )
    from verl.workers.reward_manager import (  # pyright: ignore[reportMissingImports]
        NaiveRewardManager,
    )
except ModuleNotFoundError:
    verl = None
    ray = None


from oumi.core.configs import DatasetSplitParams, TrainingConfig
from oumi.core.processors.base_processor import BaseProcessor
from oumi.core.tokenizers import BaseTokenizer
from oumi.core.trainers.base_trainer import BaseTrainer
from oumi.utils.logging import logger
from oumi.utils.verl_model_merger import FSDPModelMerger, ModelMergerConfig

# Dataset processing function type. This function takes the following arguments:
# 1. a dataset sample.
# 2. index of the sample.
# 3. data source name
# 4. split name (train, validation, etc.)
# Returns an example converted to verl format.
_DatasetProcessFn = Callable[[dict, int, str, str], dict]


class VerlGrpoTrainer(BaseTrainer):
    """verl GRPO Trainer.

    This class wraps verl's RayPPOTrainer. This class' name is misleading as it supports
    other RL algorithms as well, including GRPO, which we use here.

    For documentation on the underlying verl RayPPOTrainer, see
    https://verl.readthedocs.io/en/latest/examples/config.html.
    """

    def __init__(
        self,
        processing_class: Optional[BaseTokenizer],
        config: TrainingConfig,
        reward_funcs: list[Callable],
        train_dataset: Dataset,
        eval_dataset: Dataset,
        processor: Optional[BaseProcessor] = None,
        cache_dir: Optional[Union[str, Path]] = None,
        **kwargs,
    ):
        """Initializes the verl trainer.

        Args:
            processing_class: The tokenizer for the model.
            config: Training config.
            reward_funcs: List of reward functions to use.
            train_dataset: Training dataset.
            eval_dataset: Validation dataset. This is required by verl.
            processor: Optional processor for the dataset. Required for VLM-s.
            cache_dir: Directory to cache verl Parquet datasets.
            **kwargs: Additional keyword arguments.
        """
        if verl is None:
            raise RuntimeError(
                "verl is not installed. "
                "Please install it with 'pip install `oumi[gpu]`'."
            )
        logger.warning(
            "VerlGrpoTrainer is experimental, and the interface is subject to change."
        )
        self._processing_class = processing_class
        self._oumi_config = copy.deepcopy(config)
        self._final_output_dir: Optional[Path] = (
            Path(self._oumi_config.training.output_dir).absolute().resolve()
            if self._oumi_config.training.output_dir
            else None
        )
        self._temp_output_dir: Optional[Path] = (
            self._final_output_dir / "verl_output" if self._final_output_dir else None
        )

        if not self._final_output_dir and config.training.save_final_model:
            raise ValueError(
                "Output directory must be specified when saving final model is enabled."
            )

        # TODO: OPE-1192 - Support multiple reward functions.
        if len(reward_funcs) > 1:
            raise ValueError("We only support up to one reward function.")
        self._reward_funcs = reward_funcs

        self._cache_dir: Path = (
            Path(cache_dir)
            if cache_dir
            else Path.home() / ".cache" / "oumi" / "verl_datasets"
        )
        self._train_dataset = train_dataset
        self._eval_dataset = eval_dataset
        # verl trainer uses private methods and properties of `transformers`
        # processor, so we need to pass the raw processor here.
        self._processor = processor.raw_processor if processor is not None else None
        # Detect what dataset post-processing function to use (if any).
        process_fn = self._detect_dataset_process_fn()
        # Generate files and set self._train_filepath and self._val_filepath.
        self._create_dataset_files(process_fn)
        self._setup_verl_trainer()

    def _detect_dataset_process_fn(
        self,
    ) -> Optional[_DatasetProcessFn]:
        """Returns a post-processing function to convert data to verl format.

        Examines dataset samples to determine what post-processing function to use.

        Returns:
            A post-processing function to convert data to verl format.
            If no post-processing is needed, returns `None`.
        """
        first_train_sample = next(iter(self._train_dataset))
        first_eval_sample = next(iter(self._eval_dataset))

        if not isinstance(first_train_sample, dict):
            raise ValueError(
                "Element type of training dataset must be a dictionary. "
                f"Got {type(first_train_sample)} instead."
            )
        if not isinstance(first_eval_sample, dict):
            raise ValueError(
                "Element type of validation dataset must be a dictionary. "
                f"Got {type(first_eval_sample)} instead."
            )

        # Detect datasets containing Conversation-s.
        if "conversation_json" in first_train_sample:
            if "conversation_json" not in first_eval_sample:
                raise ValueError(
                    "Training and validation datasets must both have the same key: "
                    "'conversation_json'."
                )
            try:
                # Check if the conversation_json is valid.
                _ = Conversation.from_json(first_train_sample["conversation_json"])
                _ = Conversation.from_json(first_eval_sample["conversation_json"])
            except Exception as e:
                raise ValueError(
                    "Invalid conversation_json in training or validation dataset."
                ) from e
            return VerlGrpoTrainer._create_verl_data_entry_from_single_turn_conversation
        return None

    @staticmethod
    def _get_data_source_name(params: DatasetSplitParams) -> str:
        """Returns the verl data source name."""
        dataset_names = list({ds.dataset_name for ds in params.datasets})
        if len(dataset_names) != 1:
            if len(dataset_names) > 1:
                raise ValueError(
                    f"Multiple dataset names found: {dataset_names}. "
                    f"Please specify a single dataset name."
                )
            else:
                raise ValueError(
                    "No dataset names found. Please check the dataset split parameters."
                )
        return dataset_names[0]

    @staticmethod
    def _extract_question_images_answer_from_single_turn_conversation(
        example: dict,
    ) -> tuple[str, list, str]:
        """Finds question, answer, and optional images in a single-turn conversation.

        Args:
            example: A dictionary containing the conversation JSON.

        Returns:
            A tuple containing the question, images, and answer.
            The list of images is empty for text-only conversations.
        """
        prompt, images, answer = (
            extract_prompt_images_completion_from_single_turn_conversation(example)
        )

        if len(images) > 0:
            # TODO: Generalize. This only works for QwenVL 2.5, which is the only
            # VLM supported by verl as of 2025-05-15.
            if not prompt.startswith("<image>"):
                prompt = "<image>" + prompt
        return (prompt, images, answer)

    @staticmethod
    def _create_verl_data_entry_from_single_turn_conversation(
        example: dict, idx: int, data_source: str, split: str
    ) -> dict:
        prompt, images, answer = (
            VerlGrpoTrainer._extract_question_images_answer_from_single_turn_conversation(
                example
            )
        )
        data = {
            "data_source": data_source,
            "prompt": [
                {
                    "role": "user",
                    "content": prompt,
                }
            ],
            "images": images,
            "ability": "math",
            "reward_model": {"style": "rule", "ground_truth": answer},
            "extra_info": {
                "split": split,
                "index": idx,
                "answer": answer,
                "question": prompt,  # TODO: extract problem
            },
        }
        return data

    def _create_dataset_files(
        self, process_fn: Optional[_DatasetProcessFn] = None
    ) -> None:
        """Creates dataset files for verl in Parquet format.

        The Parquet files are saved to the Oumi cache directory.

        Args:
            process_fn: Optional function to convert the dataset samples to verl format.
        """
        train_file = self._cache_dir / "train.parquet"
        train_dataset = self._train_dataset

        # Limit the max number of sub-processes to 8 to avoid overloading the system
        # with too many processes.
        # TODO: Make this configurable.
        num_proc = min(8, os.cpu_count() or 1)

        if process_fn is not None:
            train_data_source = self._get_data_source_name(self._oumi_config.data.train)
            train_dataset = train_dataset.map(
                function=lambda example, idx: process_fn(
                    example,
                    idx,
                    train_data_source,
                    "train",
                ),
                with_indices=True,
                num_proc=num_proc,
            )

        train_dataset.to_parquet(train_file)
        self._train_filepath = str(train_file)

        val_file = self._cache_dir / "val.parquet"
        eval_dataset = self._eval_dataset
        if process_fn is not None:
            validation_data_source = self._get_data_source_name(
                self._oumi_config.data.validation
            )
            eval_dataset = eval_dataset.map(
                function=lambda example, idx: process_fn(
                    example,
                    idx,
                    validation_data_source,
                    "validation",
                ),
                with_indices=True,
                num_proc=num_proc,
            )
        eval_dataset.to_parquet(val_file)
        self._val_filepath = str(val_file)

    def _create_config(self) -> DictConfig:
        """Creates a verl config."""
        model_params = self._oumi_config.model
        model_name = model_params.model_name

        # 1. Read verl default dict config from YAML.
        yaml_path = Path(__file__).parent / "verl_trainer_config.yaml"
        config = OmegaConf.load(yaml_path)
        config = cast(DictConfig, config)

        # 2. Set config values, ex. from Oumi config values
        config.algorithm.adv_estimator = "grpo"
        config.data.train_files = self._train_filepath
        config.data.val_files = self._val_filepath

        grpo_params = self._oumi_config.training.grpo
        training_params = self._oumi_config.training

        config.data.max_response_length = grpo_params.max_completion_length
        config.actor_rollout_ref.model.path = model_name
        config.actor_rollout_ref.actor.optim.lr = training_params.learning_rate
        config.actor_rollout_ref.model.enable_gradient_checkpointing = (
            training_params.enable_gradient_checkpointing
        )
        if grpo_params.use_vllm:
            config.actor_rollout_ref.rollout.name = "vllm"
        else:
            config.actor_rollout_ref.rollout.name = "hf"
        config.actor_rollout_ref.rollout.temperature = grpo_params.temperature
        config.actor_rollout_ref.rollout.gpu_memory_utilization = (
            grpo_params.vllm_gpu_memory_utilization
        )

        # Normally, training steps is determined by the number of epochs.
        # If max_steps is set, it will override this.
        config.trainer.total_epochs = training_params.num_train_epochs
        if training_params.max_steps != -1:
            config.trainer.total_training_steps = training_params.max_steps

        if training_params.eval_strategy == "steps":
            config.trainer.test_freq = training_params.eval_steps
        if not training_params.save_epoch:
            config.trainer.save_freq = training_params.save_steps

        # Specific checkpoint to resume from takes precedence over starting
        # from last checkpoint.
        if training_params.resume_from_checkpoint:
            config.trainer.resume_mode = "resume_path"
            config.trainer.resume_from_path = training_params.resume_from_checkpoint
        elif training_params.try_resume_from_last_checkpoint:
            config.trainer.resume_mode = "auto"

        config.trainer.logger = []
        if training_params.logging_strategy != "no":
            config.trainer.logger.append("console")
        if training_params.enable_wandb:
            config.trainer.logger.append("wandb")
        config.trainer.project_name = os.environ.get("WANDB_PROJECT", "oumi_verl")
        config.trainer.experiment_name = training_params.run_name
        config.trainer.default_local_dir = str(self._temp_output_dir or "")

        # 3. Apply user overrides
        overrides_config = OmegaConf.create(training_params.verl_config_overrides)
        config = cast(DictConfig, OmegaConf.merge(config, overrides_config))

        # 4. Finalize and validate config.

        # Resolves the value of all interpolation fields in the config.
        # ex. `prompt_length: ${data.max_prompt_length}`
        # https://omegaconf.readthedocs.io/en/2.3_branch/usage.html#omegaconf-resolve
        OmegaConf.resolve(config)

        if (
            config.actor_rollout_ref.actor.strategy == "fsdp"
            and config.actor_rollout_ref.actor.strategy != config.critic.strategy
        ):
            raise ValueError(
                "Actor and critic must use the same strategy when using FSDP."
            )
        return config

    def _setup_verl_trainer(self):
        """Sets up verl's RayPPOTrainer."""
        if ray is None:
            raise RuntimeError(
                "ray is not installed. "
                "Please install it with 'pip install `oumi[gpu]`'."
            )
        self._verl_config = self._create_config()
        logger.info(f"verl config: {pformat(self._verl_config)}")

        tokenizer = self._processing_class

        role_worker_mapping = {
            Role.ActorRollout: ray.remote(ActorRolloutRefWorker),
            Role.Critic: ray.remote(CriticWorker),
            Role.RefPolicy: ray.remote(ActorRolloutRefWorker),
        }

        # Create resource pool manager
        global_pool_id = "global_pool"
        resource_pool_spec = {
            global_pool_id: [self._verl_config.trainer.n_gpus_per_node]
            * self._verl_config.trainer.nnodes,
        }
        mapping = {
            Role.ActorRollout: global_pool_id,
            Role.Critic: global_pool_id,
            Role.RefPolicy: global_pool_id,
        }

        resource_pool_manager = ResourcePoolManager(
            resource_pool_spec=resource_pool_spec, mapping=mapping
        )

        # Create reward function manager
        compute_score = self._reward_funcs[0] if len(self._reward_funcs) > 0 else None
        reward_fn = NaiveRewardManager(
            tokenizer=tokenizer, num_examine=0, compute_score=compute_score
        )
        # num_examine=1 means to print 1 example per batch for analysis.
        val_reward_fn = NaiveRewardManager(
            tokenizer=tokenizer, num_examine=1, compute_score=compute_score
        )

        self._verl_trainer = RayPPOTrainer(
            config=self._verl_config,
            tokenizer=tokenizer,
            processor=self._processor,
            role_worker_mapping=role_worker_mapping,
            resource_pool_manager=resource_pool_manager,
            reward_fn=reward_fn,
            val_reward_fn=val_reward_fn,
        )

    def train(self) -> None:
        """Trains the model using verl's RayPPOTrainer."""
        logger.info("Initializing verl trainer workers...")
        self._verl_trainer.init_workers()
        logger.info("Starting verl training...")
        self._verl_trainer.fit()

    # TODO: OPE-1192 - Implement saving model/trainer state. verl training should
    # already handle saving models, including the final checkpoint.

    def save_state(self) -> None:
        """Saves the training state."""
        pass

    def save_model(self, config: TrainingConfig, final: bool = True) -> None:
        """Saves the model.

        Args:
            config: The Oumi training config.
            final: Whether this is the final model being saved during training.
        """
        if final:
            self._export_hf_model()

    def _export_hf_model(self) -> bool:
        """Exports the tuned model to HF format.

        This method is called after training is complete.

        Returns:
            True if the model is exported successfully, False otherwise.
        """
        if not (self._final_output_dir and self._temp_output_dir):
            return False
        final_dir = Path(self._final_output_dir)
        temp_dir = Path(self._temp_output_dir)
        all_checkpoint_dirs: list[Path] = [
            f.absolute()
            for f in temp_dir.iterdir()
            if f.is_dir()
            and f.name.startswith("global_step_")
            and (f / "actor").exists()
            and (f / "actor").is_dir()
        ]

        # Find sub-directory named `global_step_NNN` with the largest NNN.
        latest_checkpoint_step = -1
        latest_checkpoint_dir: Optional[Path] = None
        for d in all_checkpoint_dirs:
            step_str = str(d.name.removeprefix("global_step_"))
            try:
                step = int(step_str)
            except Exception as e:
                raise RuntimeError(f"Failed to extract step number from {d}") from e
            if step > latest_checkpoint_step:
                latest_checkpoint_dir = d
                latest_checkpoint_step = step

        if not latest_checkpoint_dir:
            logger.warning(f"No checkpoints found under {temp_dir}")
            return False

        logger.info(
            f"Merging and exporting model from '{latest_checkpoint_dir}' "
            f"to '{final_dir}' ..."
        )

        config = ModelMergerConfig(
            operation="merge",
            backend="fsdp",
            # TODO: Detect if tie-word-embedding is enabled, or add a config parameter.
            tie_word_embedding=False,
            local_dir=str(latest_checkpoint_dir / "actor"),
            hf_model_config_path=str(latest_checkpoint_dir / "actor" / "huggingface"),
            target_dir=str(final_dir),
        )
        merger = FSDPModelMerger(config)
        merger.merge_and_save()
        logger.info(f"Successfully exported model to '{final_dir}'!")
        return True

    def get_last_eval_metrics(self) -> dict[str, Any]:
        """TODO: This should be implemented in the future to enable verl grpo in the hyper parameters tuning loops."""  # noqa: E501
        raise NotImplementedError



import os
import unittest

import pytest

from oumi.core.configs.judge_config import JudgeConfig
from oumi.judges.simple_judge import SimpleJudge

skip_if_no_openai_key = pytest.mark.skipif(
    os.getenv("OPENAI_API_KEY") is None, reason="OPENAI_API_KEY not set"
)

YAML_JUDGE_CONFIG_XML_ENUM = """
judge_params:
    prompt_template: Is the following statement correct? {statement}
    response_format: XML
    judgment_type: ENUM
    judgment_scores:
        "Correct": 0.99
        "Unsure": 0.5
        "Incorrect": 0.01
    include_explanation: True
inference_config:
    model:
        model_name: "gpt-4.1"
    engine: OPENAI
    generation:
        max_new_tokens: 8192
        temperature: 0.0
"""

YAML_JUDGE_CONFIG_JSON_BOOL = """
judge_params:
    prompt_template: "{prompt_question} {statement}"
    template_variables:
        prompt_question: "Is the following statement correct?"
    response_format: JSON
    judgment_type: BOOL
    include_explanation: False
inference_config:
    model:
        model_name: "gpt-4.1"
    engine: OPENAI
    generation:
        max_new_tokens: 8192
        temperature: 0.0
"""

JUDGE_DATASET = [
    {"statement": "The capital of France is Paris.", "useless_field": "Not used"},
    {"statement": "The Earth is flat.", "useless_field": "Not used"},
]


@skip_if_no_openai_key
def test_simple_judge_xml_enum():
    # Instantiate the judge using a YAML configuration.
    judge_config = JudgeConfig.from_str(YAML_JUDGE_CONFIG_XML_ENUM)
    simple_judge = SimpleJudge(judge_config=judge_config)

    # Call the judge with the dataset.
    judge_output = simple_judge.judge(inputs=JUDGE_DATASET)

    # Ensure the output is correct.
    print(judge_output)
    assert len(judge_output) == 2

    assert set(judge_output[0].parsed_output.keys()) == {"judgment", "explanation"}
    assert judge_output[0].parsed_output["judgment"] == "Correct"
    assert judge_output[0].parsed_output["explanation"] is not None
    assert judge_output[0].field_values["judgment"] == "Correct"
    assert judge_output[0].field_values["explanation"] is not None
    assert judge_output[0].field_scores["judgment"] == 0.99
    assert judge_output[0].field_scores["explanation"] is None

    assert set(judge_output[1].parsed_output.keys()) == {"judgment", "explanation"}
    assert judge_output[1].parsed_output["judgment"] == "Incorrect"
    assert judge_output[1].parsed_output["explanation"] is not None
    assert judge_output[1].field_values["judgment"] == "Incorrect"
    assert judge_output[1].field_values["explanation"] is not None
    assert judge_output[1].field_scores["judgment"] == 0.01
    assert judge_output[1].field_scores["explanation"] is None


@skip_if_no_openai_key
def test_simple_judge_json_bool():
    # Instantiate the judge using a YAML configuration.
    judge_config = JudgeConfig.from_str(YAML_JUDGE_CONFIG_JSON_BOOL)
    simple_judge = SimpleJudge(judge_config=judge_config)

    # Call the judge with the dataset.
    judge_output = simple_judge.judge(inputs=JUDGE_DATASET)

    # Ensure the output is correct.
    print(judge_output)
    assert len(judge_output) == 2

    assert set(judge_output[0].parsed_output.keys()) == {"judgment"}
    assert judge_output[0].parsed_output["judgment"] == "Yes"
    assert isinstance(judge_output[0].field_values["judgment"], bool)
    assert judge_output[0].field_values["judgment"]
    assert judge_output[0].field_scores["judgment"] == 1.0

    assert set(judge_output[1].parsed_output.keys()) == {"judgment"}
    assert judge_output[1].parsed_output["judgment"] == "No"
    assert isinstance(judge_output[1].field_values["judgment"], bool)
    assert not judge_output[1].field_values["judgment"]
    assert judge_output[1].field_scores["judgment"] == 0.0


if __name__ == "__main__":
    unittest.main()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional

from oumi.core.configs import TrainingParams
from oumi.core.registry import REGISTRY


def build_metrics_function(config: TrainingParams) -> Optional[Callable]:
    """Builds the metrics function."""
    metrics_function = None
    if config.metrics_function:
        metrics_function = REGISTRY.get_metrics_function(config.metrics_function)
        if not metrics_function:
            raise KeyError(
                f"metrics_function `{config.metrics_function}` "
                "was not found in the registry."
            )

    return metrics_function



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from oumi.core.configs import JobConfig


class JobState(Enum):
    """Enum to hold the state of a job."""

    PENDING = "pending"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELLED = "cancelled"


@dataclass
class JobStatus:
    """Dataclass to hold the status of a job."""

    #: The display name of the job.
    name: str

    #: The unique identifier of the job on the cluster
    id: str

    #: The status of the job.
    status: str

    #: The cluster to which the job belongs.
    cluster: str

    #: Miscellaneous metadata about the job.
    metadata: str

    #: A flag indicating whether the job is done.
    #: True only if the job is in a terminal state (e.g. completed, failed, or
    #: canceled).
    done: bool

    #: The state of the job.
    #: For more fine-grained information about the job, see the status field.
    state: JobState


class BaseCluster(ABC):
    """Base class for a compute cluster (job queue)."""

    @abstractmethod
    def name(self) -> str:
        """Gets the name of the cluster."""
        raise NotImplementedError

    @abstractmethod
    def get_job(self, job_id: str) -> JobStatus:
        """Gets the job on this cluster if it exists, else returns None."""
        raise NotImplementedError

    @abstractmethod
    def get_jobs(self) -> list[JobStatus]:
        """Lists the jobs on this cluster."""
        raise NotImplementedError

    @abstractmethod
    def cancel_job(self, job_id: str) -> JobStatus:
        """Cancels the specified job on this cluster."""
        raise NotImplementedError

    @abstractmethod
    def run_job(self, job: JobConfig) -> JobStatus:
        """Runs the specified job on this cluster."""
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """Stops the current cluster."""
        raise NotImplementedError

    @abstractmethod
    def down(self) -> None:
        """Tears down the current cluster."""
        raise NotImplementedError

    @abstractmethod
    def get_logs_stream(
        self, cluster_name: str, job_id: Optional[str] = None
    ) -> io.TextIOBase:
        """Gets a stream that tails the logs of the target job.

        Args:
            cluster_name: The name of the cluster the job was run in.
            job_id: The ID of the job to tail the logs of. If unspecified, the most
                recent job will be used.
        """
        raise NotImplementedError



from unittest.mock import Mock, patch

import pytest

from oumi.core.configs import (
    DataParams,
    DatasetParams,
    DatasetSplitParams,
    ModelParams,
    TrainingConfig,
    TrainingParams,
)
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.train import train


@pytest.fixture
def mock_tokenizer():
    """Mock tokenizer with required properties."""
    mock_tokenizer = Mock(spec=BaseTokenizer)
    mock_tokenizer.padding_side = "right"
    mock_tokenizer.pad_token_id = 0
    mock_tokenizer.model_max_length = 2048
    return mock_tokenizer


@pytest.fixture
def base_training_config():
    """Base training config without debug flag set."""
    return TrainingConfig(
        data=DataParams(
            train=DatasetSplitParams(
                collator_name="text_with_padding",
                datasets=[DatasetParams(dataset_name="dummy", split="train")],
            )
        ),
        model=ModelParams(
            model_name="gpt2", tokenizer_name="gpt2", model_max_length=1024
        ),
        training=TrainingParams(
            log_examples=False,  # Default value
        ),
    )


@pytest.fixture
def debug_training_config(base_training_config):
    """Training config with debug flag enabled."""
    config = base_training_config
    config.training.log_examples = True
    return config


@pytest.fixture
def mock_log_params():
    """Mock for log_number_of_model_parameters function."""
    with patch("oumi.train.log_number_of_model_parameters") as mock:
        yield mock


@pytest.fixture
def mock_build_collator():
    """Mock for build_collator_from_config function."""
    with patch("oumi.train.build_collator_from_config") as mock:
        yield mock


@pytest.fixture
def mock_build_model():
    """Mock for build_model function."""
    with patch("oumi.train.build_model") as mock:
        yield mock


@pytest.fixture
def mock_build_dataset_mixture():
    """Mock for build_dataset_mixture function."""
    with patch("oumi.train.build_dataset_mixture") as mock:
        yield mock


@pytest.fixture
def mock_build_trainer():
    """Mock for build_trainer function."""
    with patch("oumi.train.build_trainer") as mock:
        yield mock


@pytest.fixture
def mock_configure_logger():
    """Mock for configure_logger function."""
    with patch("oumi.train.configure_logger") as mock:
        yield mock


def test_train_function_passes_debug_flag_correctly(
    mock_build_trainer,
    mock_build_dataset_mixture,
    mock_build_model,
    mock_build_collator,
    mock_log_params,
    mock_configure_logger,
    debug_training_config,
):
    """Test that train function passes log_examples to build_collator_from_config."""

    # Setup mocks to avoid actual training
    mock_model = Mock()
    mock_model.named_modules.return_value = []
    mock_tokenizer = Mock()
    mock_model_tuple = (mock_model, mock_tokenizer)
    mock_build_model.return_value = mock_model_tuple

    mock_dataset = Mock()
    mock_build_dataset_mixture.return_value = mock_dataset

    mock_collator = Mock()
    mock_build_collator.return_value = mock_collator

    mock_trainer = Mock()
    mock_build_trainer.return_value = mock_trainer
    mock_trainer.train.return_value = Mock()

    # Call train function
    train(debug_training_config)

    # Verify that build_collator_from_config was called with the correct debug flag
    mock_build_collator.assert_called_once()
    call_args = mock_build_collator.call_args

    # The debug parameter should match config.training.log_examples
    assert call_args.args[0] is debug_training_config
    assert call_args.kwargs["debug"] is True

    # Verify the debug flag came from the config
    assert debug_training_config.training.log_examples is True



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lambda AI inference engine implementation."""

from typing import Optional

from typing_extensions import override

from oumi.core.configs import RemoteParams
from oumi.inference.remote_inference_engine import RemoteInferenceEngine


class LambdaInferenceEngine(RemoteInferenceEngine):
    """Engine for running inference against the Lambda AI API.

    This class extends RemoteInferenceEngine to provide specific functionality
    for interacting with Lambda AI's language models via their API. It handles
    the conversion of Oumi's Conversation objects to Lambda AI's expected input
    format, as well as parsing the API responses back into Conversation objects.
    """

    @property
    @override
    def base_url(self) -> Optional[str]:
        """Return the default base URL for the Lambda AI API."""
        return "https://api.lambda.ai/v1/chat/completions"

    @property
    @override
    def api_key_env_varname(self) -> Optional[str]:
        """Return the default environment variable name for the Lambda AI API key."""
        return "LAMBDA_API_KEY"

    @override
    def _default_remote_params(self) -> RemoteParams:
        """Returns the default remote parameters."""
        return RemoteParams(num_workers=20, politeness_policy=60.0)



import functools
import itertools
import json
from typing import Optional
from unittest.mock import patch

import PIL.Image
import pydantic
import pytest

from oumi.core.configs import (
    GenerationParams,
    GuidedDecodingParams,
    ModelParams,
    RemoteParams,
)
from oumi.core.types.conversation import (
    ContentItem,
    Conversation,
    Message,
    Role,
    Type,
)
from oumi.inference.sglang_inference_engine import SGLangInferenceEngine
from oumi.utils.image_utils import (
    create_png_bytes_from_image,
)
from oumi.utils.logging import logger


class SamplePydanticType(pydantic.BaseModel):
    name: str
    score: float


def create_test_remote_params():
    return RemoteParams(api_key="test_api_key", api_url="<placeholder>")


def create_test_vision_language_engine() -> SGLangInferenceEngine:
    return SGLangInferenceEngine(
        model_params=ModelParams(
            model_name="llava-hf/llava-1.5-7b-hf",
            torch_dtype_str="bfloat16",
            model_max_length=1024,
            chat_template="llama3-instruct",
            trust_remote_code=True,
        ),
        remote_params=create_test_remote_params(),
    )


def create_test_text_only_engine() -> SGLangInferenceEngine:
    return SGLangInferenceEngine(
        model_params=ModelParams(
            model_name="openai-community/gpt2",
            torch_dtype_str="bfloat16",
            model_max_length=1024,
            chat_template="llama3-instruct",
            trust_remote_code=True,
            tokenizer_pad_token="<|endoftext|>",
        ),
        remote_params=create_test_remote_params(),
    )


@functools.cache
def _generate_all_engines() -> list[SGLangInferenceEngine]:
    return [create_test_vision_language_engine(), create_test_text_only_engine()]


@pytest.mark.parametrize(
    "engine,guided_decoding,num_images",
    list(
        itertools.product(
            _generate_all_engines(),
            [
                None,
                GuidedDecodingParams(choice=["apple", "pear"]),
                GuidedDecodingParams(json={"enum": ["apple", "pear"]}),
                GuidedDecodingParams(json=SamplePydanticType(name="hey", score=0.7)),
                GuidedDecodingParams(json=SamplePydanticType),
                GuidedDecodingParams(regex="(apple|pear)"),
            ],
            [None],  # num_images
        )
    )
    + [
        (create_test_vision_language_engine(), None, 2),
    ],
)
def test_convert_conversation_to_api_input(
    engine: SGLangInferenceEngine,
    guided_decoding: Optional[GuidedDecodingParams],
    num_images: Optional[int],
):
    is_vision_language: bool = "llava" in engine._model_params.model_name.lower()
    num_images = num_images or (1 if is_vision_language else 0)

    pil_image = PIL.Image.new(mode="RGB", size=(32, 48))
    png_bytes = create_png_bytes_from_image(pil_image)
    conversation = Conversation(
        messages=(
            [Message(role=Role.SYSTEM, content="System message")]
            + (
                [
                    Message(
                        role=Role.USER,
                        content=(
                            [ContentItem(binary=png_bytes, type=Type.IMAGE_BINARY)]
                            * num_images
                            + [ContentItem(type=Type.TEXT, content="User message")]
                        ),
                    )
                ]
                if is_vision_language
                else [Message(role=Role.USER, content="User message")]
            )
            + [
                Message(role=Role.ASSISTANT, content="Assistant message"),
            ]
        ),
        metadata={"key": "value"},
        conversation_id="test_id",
    )
    generation_params = GenerationParams(
        max_new_tokens=100,
        temperature=0.2,
        top_p=0.8,
        min_p=0.1,
        frequency_penalty=0.3,
        presence_penalty=0.4,
        stop_strings=["stop it"],
        stop_token_ids=[32000],
        guided_decoding=guided_decoding,
    )

    if num_images > 1 and is_vision_language and not engine._supports_multiple_images:
        with pytest.raises(ValueError, match="A conversation contains too many images"):
            engine._convert_conversation_to_api_input(
                conversation, generation_params, ModelParams()
            )

        return

    result = engine._convert_conversation_to_api_input(
        conversation, generation_params, ModelParams()
    )

    assert isinstance(engine._tokenizer.bos_token, str), (
        "bos_token: {engine._tokenizer.bos_token}"
    )
    expected_prompt = (
        "\n\n".join(
            [
                (
                    engine._tokenizer.bos_token
                    + "<|start_header_id|>system<|end_header_id|>"
                ),
                "System message<|eot_id|><|start_header_id|>user<|end_header_id|>",
            ]
            + [
                (
                    ("<|image|>" if is_vision_language else "")
                    + "User message<|eot_id|>"
                    + "<|start_header_id|>assistant<|end_header_id|>"
                ),
                (
                    "Assistant message<|eot_id|><|start_header_id|>assistant"
                    "<|end_header_id|>"
                ),
            ]
        )
        + "\n\n"
    )

    assert "text" in result, result
    logger.info(f"result['text']:\n{result['text']}\n\n")
    logger.info(f"expected_prompt:\n{expected_prompt}\n\n")
    assert result["text"] == expected_prompt, result
    if is_vision_language:
        assert num_images >= 1
        assert "image_data" in result, result
        if num_images > 1:
            assert isinstance(result["image_data"], list)
            assert len(result["image_data"]) == num_images
            for idx in range(num_images):
                assert result["image_data"][idx].startswith("data:image/png;base64,"), (
                    result
                )

        else:
            assert isinstance(result["image_data"], str)
            assert result["image_data"].startswith("data:image/png;base64,"), result
    else:
        assert "image_data" not in result, result
    assert "sampling_params" in result, result
    assert result["sampling_params"]["max_new_tokens"] == 100, result
    assert result["sampling_params"]["temperature"] == 0.2, result
    assert result["sampling_params"]["top_p"] == 0.8, result
    assert result["sampling_params"]["min_p"] == 0.1, result
    assert result["sampling_params"]["frequency_penalty"] == 0.3, result
    assert result["sampling_params"]["presence_penalty"] == 0.4, result
    assert result["sampling_params"]["stop"] == ["stop it"], result
    assert result["sampling_params"]["stop_token_ids"] == [32000], result

    expect_valid_regex = False
    expect_valid_json_schema = False
    if guided_decoding is not None:
        if guided_decoding.regex is not None:
            assert "regex" in result["sampling_params"]
            assert result["sampling_params"]["regex"] == guided_decoding.regex
            expect_valid_regex = True
        elif guided_decoding.json is not None:
            assert "json_schema" in result["sampling_params"]
            if isinstance(guided_decoding.json, pydantic.BaseModel) or (
                isinstance(guided_decoding.json, type)
                and issubclass(guided_decoding.json, pydantic.BaseModel)
            ):
                assert (
                    json.loads(result["sampling_params"]["json_schema"])
                    == guided_decoding.json.model_json_schema()
                )

            else:
                assert (
                    json.loads(result["sampling_params"]["json_schema"])
                    == guided_decoding.json
                )
            expect_valid_json_schema = True
        elif guided_decoding.choice is not None:
            assert "json_schema" in result["sampling_params"]
            assert json.loads(result["sampling_params"]["json_schema"]) == {
                "enum": list(guided_decoding.choice)
            }
            expect_valid_json_schema = True

    if not expect_valid_regex:
        assert "regex" in result["sampling_params"]
        assert result["sampling_params"]["regex"] is None
    if not expect_valid_json_schema:
        assert "json_schema" in result["sampling_params"]
        assert result["sampling_params"]["json_schema"] is None


@pytest.mark.parametrize(
    "engine",
    _generate_all_engines(),
)
def test_convert_api_output_to_conversation(engine):
    original_conversation = Conversation(
        messages=[
            Message(content="User message", role=Role.USER),
        ],
        metadata={"key": "value"},
        conversation_id="test_id",
    )
    api_response = {"text": "Assistant response"}

    result = engine._convert_api_output_to_conversation(
        api_response, original_conversation
    )

    assert len(result.messages) == 2
    assert result.messages[0].content == "User message"
    assert result.messages[1].content == "Assistant response"
    assert result.messages[1].role == Role.ASSISTANT
    assert result.metadata == {"key": "value"}
    assert result.conversation_id == "test_id"


@pytest.mark.parametrize(
    "engine",
    _generate_all_engines(),
)
def test_get_request_headers(engine):
    remote_params = RemoteParams(api_key="test_api_key", api_url="<placeholder>")

    with patch.object(
        SGLangInferenceEngine,
        "_get_api_key",
        return_value="test_api_key",
    ):
        result = engine._get_request_headers(remote_params)

    assert result["Content-Type"] == "application/json"


@pytest.mark.parametrize(
    "engine",
    _generate_all_engines(),
)
def test_get_supported_params(engine):
    assert engine.get_supported_params() == set(
        {
            "frequency_penalty",
            "guided_decoding",
            "max_new_tokens",
            "min_p",
            "presence_penalty",
            "stop_strings",
            "stop_token_ids",
            "temperature",
            "top_p",
        }
    )



import argparse
import time

import torch
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP

from oumi.core.distributed import (
    barrier,
    cleanup_distributed,
    get_device_rank_info,
    init_distributed,
    is_distributed,
    is_local_process_zero,
)
from oumi.models.mlp import MLPEncoder
from oumi.utils.logging import logger, update_logger_level


def benchmark_barrier(device_info, num_iterations: int = 1):
    """Benchmarks the time taken to execute a barrier operation."""
    start_time = time.perf_counter()
    sleep_time_seconds = 1
    for _ in range(num_iterations):
        barrier()
        time.sleep(sleep_time_seconds)
    end_time = time.perf_counter()

    total_time_seconds = end_time - start_time
    total_barrier_time_seconds = (
        total_time_seconds - sleep_time_seconds * num_iterations
    )
    avg_time = total_barrier_time_seconds / num_iterations

    if is_local_process_zero():
        logger.info(f"Average barrier time: {avg_time * 1000:.3f} ms")


def benchmark_ddp(device_info, num_iterations: int = 1):
    """Benchmarks the time taken to execute a forward and backward pass with DDP."""
    rank = device_info.local_rank
    device = f"cuda:{rank}"

    model = MLPEncoder().to(device)
    ddp_model = DDP(model, device_ids=[rank])

    optimizer = torch.optim.SGD(ddp_model.parameters(), lr=0.01)

    forward_times = []
    backward_times = []

    for _ in range(num_iterations):
        input_tensor = torch.randint(high=768, size=(1000,)).to(rank)
        target_tensor = torch.randint(high=10, size=(1000,)).to(rank)

        # Forward pass
        forward_start = time.perf_counter()
        output = ddp_model(input_ids=input_tensor, labels=target_tensor)
        loss = output["loss"].sum()

        forward_end = time.perf_counter()
        forward_times.append(forward_end - forward_start)

        # Backward pass
        backward_start = time.perf_counter()
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()
        backward_end = time.perf_counter()
        backward_times.append(backward_end - backward_start)

        torch.cuda.synchronize()

    avg_forward = sum(forward_times) / num_iterations
    avg_backward = sum(backward_times) / num_iterations

    if is_local_process_zero():
        logger.info(f"Average DDP forward time: {avg_forward * 1000:.3f} ms")
        logger.info(f"Average DDP backward time: {avg_backward * 1000:.3f} ms")


def benchmark_all_reduce(device_info, num_iterations: int = 1, tensor_size=1000000):
    """Benchmarks the time taken to execute an all-reduce operation."""
    tensor = torch.randn(tensor_size).to(device_info.local_rank)

    start_time = time.time()
    for _ in range(num_iterations):
        dist.all_reduce(tensor, op=dist.ReduceOp.SUM)
        torch.cuda.synchronize()
    end_time = time.time()

    avg_time = (end_time - start_time) / num_iterations

    if is_local_process_zero():
        logger.info(f"Average all_reduce time: {avg_time * 1000:.3f} ms")


def main(args):
    """Main function for the benchmark script."""
    if is_distributed():
        logger.info("Benchmarking in distributed mode...")
        init_distributed()
    else:
        logger.info("Running benchmark in single-process mode.")

    #
    # Run tests
    #
    device_info = get_device_rank_info()
    torch.cuda.set_device(device_info.local_rank)

    logger.info("Benchmarking barrier...")
    torch.cuda.synchronize()
    benchmark_barrier(device_info, num_iterations=args.num_iterations)

    logger.info("Benchmarking DDP forward/backward...")
    torch.cuda.synchronize()
    benchmark_ddp(device_info, num_iterations=args.num_iterations)

    logger.info("Benchmarking all_reduce...")
    torch.cuda.synchronize()
    benchmark_all_reduce(device_info, num_iterations=args.num_iterations)

    if is_distributed():
        cleanup_distributed()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Benchmark Distributed/NCCL operations"
    )
    parser.add_argument(
        "--num_iterations",
        type=int,
        default=10,
        help="Number of iterations for each benchmark",
    )
    parser.add_argument(
        "--log-level",
        type=str,
        default="info",
        help="Log level.",
    )
    args = parser.parse_args()

    update_logger_level("oumi", level=args.log_level)

    main(args)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.configs.params.synthesis_params import GeneralSynthesisParams
from oumi.utils.placeholders import resolve_placeholders


class _AttributeParentInfo:
    """Information about a parent of a sampled attribute."""

    def __init__(self, parent_name: str, parent_description: str):
        """Initialize the attribute parent info."""
        self._parent_name = parent_name
        self.description = parent_description

    def __str__(self) -> str:
        return self._parent_name


class _AttributeInfo:
    """Information about a sampled attribute.

    Used to format the string for a sample.

    Example:
        attribute_id: "complexity"
        parent_name: "Complexity"
        parent_description: "The complexity of the text."
        value_name: "High"
        value_description: "The text is complex."

    Formatting string:
        {complexity.parent} ({complexity.parent.description})
        {complexity} ({complexity.description})

    Result:
        Complexity (The complexity of the text.)
        High (The text is complex.)
    """

    def __init__(
        self,
        attribute_id: str,
        parent_name: str,
        parent_description: str,
        value_name: str,
        value_description: str,
    ):
        """Initialize the attribute value info."""
        self.attribute_id = attribute_id
        self._attribute_name = value_name
        self.description = value_description
        self.parent = _AttributeParentInfo(parent_name, parent_description)

    def __str__(self) -> str:
        return self._attribute_name


class AttributeFormatter:
    """Formats a sample using a format string.

    Integrates information from permutable attributes to support
    formatting of placeholders in the format string (i.e. {attribute_id.value}).
    """

    def __init__(self, params: GeneralSynthesisParams):
        """Initialize the formatter."""
        self._params = params
        self._sampled_attribute_map = (
            {
                sampled_attr.id: sampled_attr
                for sampled_attr in params.sampled_attributes
            }
            if params.sampled_attributes
            else {}
        )
        self._sampled_attribute_info = {}

        # Pre-compute the attribute info for each possible value
        for attribute_id, attribute in self._sampled_attribute_map.items():
            for value in attribute.possible_values:
                key = (attribute_id, value.id)
                self._sampled_attribute_info[key] = _AttributeInfo(
                    attribute_id=attribute_id,
                    parent_name=attribute.name,
                    parent_description=attribute.description,
                    value_name=value.name,
                    value_description=value.description,
                )

    def format(
        self,
        sample: dict[str, str],
        format_string: str,
        missing_values_allowed: bool = False,
    ) -> str:
        """Format a sample using a format string.

        Args:
            sample: The sample to format.
            format_string: The format string to use.
            missing_values_allowed: If True, missing values are allowed in the sample.

        Returns:
            The formatted string.
        """
        attr_values = {}
        for attribute_id, attribute_value in sample.items():
            if self._is_sampled_attribute(attribute_id):
                value_id = attribute_value
                attr_values[attribute_id] = self._get_sampled_attribute_value_info(
                    attribute_id, value_id
                )
            else:
                attr_values[attribute_id] = attribute_value

        formatted_string = resolve_placeholders(
            format_string,
            attr_values,
            missing_values_allowed=missing_values_allowed,
        )
        return formatted_string

    def _is_sampled_attribute(self, attribute_id: str) -> bool:
        """Check if the attribute is a sampled attribute."""
        return attribute_id in self._sampled_attribute_map

    def _get_sampled_attribute_value_info(
        self, attribute_id: str, attribute_value_id: str
    ) -> _AttributeInfo:
        """Get the string representation information for a sampled attribute."""
        key = (attribute_id, attribute_value_id)
        if key in self._sampled_attribute_info:
            return self._sampled_attribute_info[key]

        raise ValueError(
            f"Attribute value {attribute_value_id} not found for "
            f"attribute {attribute_id}"
        )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Porting the Alpaca evaluation dataset to Oumi.

For more info see: https://github.com/tatsu-lab/alpaca_eval
"""

from typing import Union, cast

import pandas as pd

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("tatsu-lab/alpaca_eval")
class AlpacaEvalDataset(BaseSftDataset):
    system_prompt_with_context = (
        "Below is an instruction that describes a task, "
        "paired with an input that provides further context. "
        "Write a response that appropriately completes the request."
    )

    system_prompt_without_context = (
        "Below is an instruction that describes a task. "
        "Write a response that appropriately completes the request."
    )

    default_dataset = "tatsu-lab/alpaca_eval"

    def __init__(
        self,
        *,
        include_system_prompt: bool = False,
        unused_entries_to_metadata: bool = False,
        trust_remote_code: bool = True,
        **kwargs,
    ) -> None:
        """Initializes a new instance of the AlpacaEvalDataset class.

        Args:
            include_system_prompt: Whether to include a system prompt in the
                conversation.
            unused_entries_to_metadata (bool): Whether to save entries that were not
                used in the conversation (entries other than `instruction`, `input`)
                as metadata.
            trust_remote_code: Whether to trust remote code.
            **kwargs: Additional keyword arguments.
        """
        self.include_system_prompt = include_system_prompt
        self.unused_entries_to_metadata = unused_entries_to_metadata

        super().__init__(**kwargs, trust_remote_code=trust_remote_code)

    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Preprocesses the inputs of the example and returns a dictionary.

        Args:
            example (dict or Pandas Series): An example containing `input` (optional),
                `instruction` entries.

        Returns:
            dict: The input example converted to Alpaca dictionary format.

        Note:
            If `unused_entries_to_metadata` is set: all example's entries, other than
            the expected ones (i.e., `input` and `instruction`), are saved as metadata.
        """
        messages = []

        # Use default Alpaca user prompt template.
        if ("input" in example) and len(example["input"]) > 0:
            # This example has both an instruction and a user input.
            user_prompt = f"{example['instruction']}\n\n### Input:\n{example['input']}"
            system_prompt = self.system_prompt_with_context
        else:
            user_prompt = cast(str, example["instruction"])
            system_prompt = self.system_prompt_without_context

        # Create message list.
        if self.include_system_prompt:
            messages.append(Message(role=Role.SYSTEM, content=system_prompt))
        messages.append(Message(role=Role.USER, content=user_prompt))

        # Retain entries (other than `instruction`, `input`) as metadata.
        metadata_fields = set()
        if self.unused_entries_to_metadata:
            if isinstance(example, pd.Series):
                metadata_fields = {str(i) for i in example.index}
            elif isinstance(example, dict):
                metadata_fields = {str(key) for key in example.keys()}
            metadata_fields = metadata_fields - {"instruction", "input"}
        metadata = {field: example[field] for field in metadata_fields}

        return Conversation(messages=messages, metadata=metadata)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OpenO1 synthetic reasoning SFT dataset."""

from oumi.core.registry.registry import register_dataset
from oumi.datasets.sft.prompt_response import PromptResponseDataset


@register_dataset("O1-OPEN/OpenO1-SFT")
class OpenO1SFTDataset(PromptResponseDataset):
    """Synthetic reasoning SFT dataset."""

    default_dataset = "O1-OPEN/OpenO1-SFT"

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """Initializes a dataset for OpenO1SFT from HuggingFace."""
        super().__init__(
            hf_dataset_path="O1-OPEN/OpenO1-SFT",
            prompt_column="instruction",
            response_column="output",
            **kwargs,
        )



"""Experimental package for the Oumi (Open Universal Machine Intelligence) library.

This package contains experimental features and components that are still
under development or testing. Contents here may change frequently and are
not guaranteed to be stable.

Warning:
   The features in this package are experimental and may be subject to
   significant changes or removal without notice. Use with caution in
   production environments.

Note:
   Feedback and contributions to experimental features are welcome.
   Please refer to the contribution guidelines for more information.

See Also:
    :mod:`oumi.core`: For stable, production-ready features
"""



from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import torch

from oumi.utils.device_utils import (
    _get_nvidia_gpu_runtime_info_impl,
    get_nvidia_gpu_fan_speeds,
    get_nvidia_gpu_memory_utilization,
    get_nvidia_gpu_power_usage,
    get_nvidia_gpu_runtime_info,
    get_nvidia_gpu_temperature,
    log_nvidia_gpu_fan_speeds,
    log_nvidia_gpu_memory_utilization,
    log_nvidia_gpu_power_usage,
    log_nvidia_gpu_runtime_info,
    log_nvidia_gpu_temperature,
)
from tests.markers import requires_cuda_initialized, requires_cuda_not_available


@requires_cuda_initialized()
def test_nvidia_gpu_memory_utilization():
    num_devices = torch.cuda.device_count()
    if num_devices > 0:
        for device_index in range(0, num_devices):
            memory_mib = get_nvidia_gpu_memory_utilization(device_index)
            assert memory_mib > 1  # Must have at least 1 MB
            assert memory_mib < 1024 * 1024  # No known GPU has 1 TB of VRAM yet.
            log_nvidia_gpu_memory_utilization(device_index)

        # Test default argument value
        assert get_nvidia_gpu_memory_utilization() == get_nvidia_gpu_memory_utilization(
            0
        )
    else:
        # Test default argument value
        assert get_nvidia_gpu_memory_utilization() == 0.0

    log_nvidia_gpu_memory_utilization()


@requires_cuda_not_available()
def test_nvidia_gpu_memory_utilization_no_cuda():
    assert get_nvidia_gpu_memory_utilization() == 0.0
    log_nvidia_gpu_memory_utilization()


@requires_cuda_initialized()
def test_nvidia_gpu_temperature():
    num_devices = torch.cuda.device_count()
    if num_devices > 0:
        for device_index in range(0, num_devices):
            temperature = get_nvidia_gpu_temperature(device_index)
            assert temperature > 0 and temperature < 100
            log_nvidia_gpu_temperature(device_index)

        # Test default argument value
        temperature = get_nvidia_gpu_temperature()
        assert temperature > 0 and temperature < 100
    else:
        # Test default argument value
        assert get_nvidia_gpu_temperature() == 0.0

    log_nvidia_gpu_temperature()


@requires_cuda_not_available()
def test_nvidia_gpu_temperature_no_cuda():
    assert get_nvidia_gpu_temperature() == 0.0
    log_nvidia_gpu_temperature()


@requires_cuda_initialized()
def test_nvidia_gpu_fan_speeds():
    num_devices = torch.cuda.device_count()
    if num_devices > 0:
        for device_index in range(0, num_devices):
            fan_speeds = get_nvidia_gpu_fan_speeds(device_index)
            if fan_speeds:
                assert len(fan_speeds) > 0
                fan_speeds = np.array(fan_speeds)
                assert np.all(fan_speeds >= 0)
                assert np.all(fan_speeds <= 100)
            else:
                assert fan_speeds == tuple()
            log_nvidia_gpu_fan_speeds(device_index)

        # Test default argument value
        fan_speeds = get_nvidia_gpu_fan_speeds()
        if fan_speeds:
            assert len(fan_speeds) > 0
            fan_speeds = np.array(fan_speeds)
            assert np.all(fan_speeds >= 0)
            assert np.all(fan_speeds <= 100)
        else:
            assert fan_speeds == tuple()
        log_nvidia_gpu_fan_speeds(device_index)
    else:
        # Test default argument value
        assert get_nvidia_gpu_fan_speeds() == tuple()

    log_nvidia_gpu_fan_speeds()


@requires_cuda_not_available()
def test_nvidia_gpu_fan_speeds_no_cuda():
    assert get_nvidia_gpu_fan_speeds() == tuple()
    log_nvidia_gpu_fan_speeds()


@requires_cuda_initialized()
def test_nvidia_gpu_power_usage():
    num_devices = torch.cuda.device_count()
    if num_devices > 0:
        for device_index in range(0, num_devices):
            watts = get_nvidia_gpu_power_usage(device_index)
            assert watts > 0 and watts < 2000
            log_nvidia_gpu_power_usage(device_index)

        # Test default argument value
        watts = get_nvidia_gpu_power_usage()
        assert watts > 0 and watts < 2000
    else:
        # Test default argument value
        assert get_nvidia_gpu_power_usage() == 0.0

    log_nvidia_gpu_power_usage()


@requires_cuda_not_available()
def test_nvidia_gpu_power_usage_no_cuda():
    assert get_nvidia_gpu_power_usage() == 0.0
    log_nvidia_gpu_power_usage()


@requires_cuda_initialized()
def test_nvidia_gpu_runtime_info():
    num_devices = torch.cuda.device_count()
    if num_devices > 0:
        for device_index in range(0, num_devices):
            info = get_nvidia_gpu_runtime_info(device_index)
            assert info is not None
            assert info.device_index == device_index
            assert info.device_count == num_devices
            assert info.used_memory_mb is not None and info.used_memory_mb > 0
            assert (
                info.temperature is not None
                and info.temperature >= 0
                and info.temperature <= 100
            )
            assert info.fan_speed is None or (
                info.fan_speed >= 0 and info.fan_speed <= 100
            )
            assert info.fan_speeds is None or len(info.fan_speeds) > 0
            assert info.power_usage_watts is not None and info.power_usage_watts >= 0
            assert info.power_limit_watts is not None and info.power_limit_watts > 0
            assert (
                info.gpu_utilization is not None
                and info.gpu_utilization >= 0
                and info.gpu_utilization <= 100
            )
            assert (
                info.memory_utilization is not None
                and info.memory_utilization >= 0
                and info.memory_utilization <= 100
            )
            assert info.performance_state is not None
            assert info.performance_state >= 0 and info.performance_state <= 32
            assert (
                info.clock_speed_graphics is not None and info.clock_speed_graphics > 0
            )
            assert info.clock_speed_sm is not None and info.clock_speed_sm > 0
            assert info.clock_speed_memory is not None and info.clock_speed_memory > 0

            log_nvidia_gpu_runtime_info(device_index)

        # Test default argument value
        info = get_nvidia_gpu_runtime_info()
        assert info is not None
        assert info.device_index == 0
        assert info.device_count == num_devices
        assert info.used_memory_mb is not None and info.used_memory_mb > 0
        assert (
            info.temperature is not None
            and info.temperature >= 0
            and info.temperature <= 100
        )
        assert info.fan_speed is None or (info.fan_speed >= 0 and info.fan_speed <= 100)
        assert info.fan_speeds is None or len(info.fan_speeds) > 0
        assert info.power_usage_watts is not None and info.power_usage_watts >= 0
        assert info.power_limit_watts is not None and info.power_limit_watts > 0
        assert (
            info.gpu_utilization is not None
            and info.gpu_utilization >= 0
            and info.gpu_utilization <= 100
        )
        assert (
            info.memory_utilization is not None
            and info.memory_utilization >= 0
            and info.memory_utilization <= 100
        )
        assert info.performance_state is not None
        assert info.performance_state >= 0 and info.performance_state <= 32
        assert info.clock_speed_graphics is not None and info.clock_speed_graphics > 0
        assert info.clock_speed_sm is not None and info.clock_speed_sm > 0
        assert info.clock_speed_memory is not None and info.clock_speed_memory > 0
    else:
        # Test default argument value
        assert get_nvidia_gpu_runtime_info() is None

    log_nvidia_gpu_runtime_info()


@requires_cuda_not_available()
def test_nvidia_gpu_runtime_info_no_cuda():
    assert get_nvidia_gpu_runtime_info() is None
    log_nvidia_gpu_runtime_info()


def _create_mock_pynvml_with_not_supported_error():
    """Helper function to create a mock pynvml with NVMLError_NotSupported."""
    mock_pynvml = MagicMock()

    # Mock successful initialization and device count
    mock_pynvml.nvmlInit.return_value = None
    mock_pynvml.nvmlDeviceGetCount.return_value = 1

    # Mock successful GPU handle retrieval by default
    mock_gpu_handle = MagicMock()
    mock_pynvml.nvmlDeviceGetHandleByIndex.return_value = mock_gpu_handle

    # Create a mock NVMLError_NotSupported exception
    class MockNVMLError_NotSupported(Exception):
        pass

    mock_pynvml.NVMLError_NotSupported = MockNVMLError_NotSupported

    return mock_pynvml, MockNVMLError_NotSupported


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "name": "utilization",
            "mock_method": "nvmlDeviceGetUtilizationRates",
            "function_kwargs": {"utilization": True},
            "expected_log_message": "GPU utilization not supported for device: 0",
            "expected_none_fields": ["gpu_utilization", "memory_utilization"],
            "should_return_none": False,
        },
        {
            "name": "memory",
            "mock_method": "nvmlDeviceGetMemoryInfo",
            "function_kwargs": {"memory": True},
            "expected_log_message": (
                "pyNVML GPU memory info not supported for device: 0"
            ),
            "expected_none_fields": ["used_memory_mb"],
            "should_return_none": False,
        },
        {
            "name": "temperature",
            "mock_method": "nvmlDeviceGetTemperature",
            "function_kwargs": {"temperature": True},
            "expected_log_message": (
                "pyNVML GPU temperature not supported for device: 0"
            ),
            "expected_none_fields": ["temperature"],
            "should_return_none": False,
        },
        {
            "name": "fan_speed",
            "mock_method": "nvmlDeviceGetFanSpeed",
            "function_kwargs": {"fan_speed": True},
            "expected_log_message": (
                "pyNVML GPU fan speed not supported for device: 0"
            ),
            "expected_none_fields": ["fan_speed", "fan_speeds"],
            "should_return_none": False,
        },
        {
            "name": "power_usage",
            "mock_method": "nvmlDeviceGetPowerUsage",
            "function_kwargs": {"power_usage": True},
            "expected_log_message": (
                "pyNVML GPU power usage not supported for device: 0"
            ),
            "expected_none_fields": ["power_usage_watts", "power_limit_watts"],
            "should_return_none": False,
        },
        {
            "name": "performance_state",
            "mock_method": "nvmlDeviceGetPerformanceState",
            "function_kwargs": {"performance_state": True},
            "expected_log_message": (
                "pyNVML GPU performance state not supported for device: 0"
            ),
            "expected_none_fields": ["performance_state"],
            "should_return_none": False,
        },
        {
            "name": "clock_speed",
            "mock_method": "nvmlDeviceGetClockInfo",
            "function_kwargs": {"clock_speed": True},
            "expected_log_message": (
                "pyNVML GPU clock speed not supported for device: 0"
            ),
            "expected_none_fields": [
                "clock_speed_graphics",
                "clock_speed_sm",
                "clock_speed_memory",
            ],
            "should_return_none": False,
        },
    ],
)
def test_debug_logging_for_unsupported_gpu_operations(caplog, test_case):
    """Test that debug logging occurs when GPU operations are not supported."""
    mock_pynvml, MockNVMLError_NotSupported = (
        _create_mock_pynvml_with_not_supported_error()
    )

    # Set up the specific method to raise NotSupported error
    getattr(
        mock_pynvml, test_case["mock_method"]
    ).side_effect = MockNVMLError_NotSupported("Not supported")

    with patch("oumi.utils.device_utils.pynvml", mock_pynvml):
        with caplog.at_level("DEBUG", logger="oumi"):
            result = _get_nvidia_gpu_runtime_info_impl(
                device_index=0, **test_case["function_kwargs"]
            )

        # Verify debug message was logged
        assert any(
            test_case["expected_log_message"] in record.message
            for record in caplog.records
            if record.levelname == "DEBUG"
        ), f"Expected debug message not found: {test_case['expected_log_message']}"

        # Check return value and field values
        if test_case["should_return_none"]:
            assert result is None
        else:
            assert result is not None
            for field_name in test_case["expected_none_fields"]:
                assert getattr(result, field_name) is None, (
                    f"Expected {field_name} to be None"
                )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import time
import uuid
from dataclasses import dataclass
from datetime import datetime
from functools import reduce
from pathlib import Path
from typing import Any, Optional

from oumi.core.configs import JobConfig
from oumi.core.launcher import BaseCluster, JobStatus
from oumi.launcher.clients.slurm_client import SlurmClient, SlurmLogStream
from oumi.utils.logging import logger

_OUMI_SLURM_CONNECTIONS = "OUMI_SLURM_CONNECTIONS"


def _format_date(date: datetime) -> str:
    """Formats the provided date as a string.

    Args:
        date: The date to format.

    Returns:
        The formatted date.
    """
    return date.strftime("%Y%m%d_%H%M%S%f")


def _last_sbatch_line(script: list[str]) -> int:
    """Finds the last SBATCH instruction line in the script.

    Args:
        script: The lines of the script.

    Returns:
        The index of the last SBATCH instruction line. -1 if not found.
    """
    return reduce(
        lambda acc, val: val[0] if val[1].startswith("#SBATCH") else acc,
        enumerate(script),
        -1,
    )


def _create_job_script(job: JobConfig) -> str:
    """Creates a job script for the specified job.

    Args:
        job: The job to create a script for.

    Returns:
        The script as a string.
    """
    setup_lines = [] if not job.setup else job.setup.strip().split("\n")
    run_lines = job.run.strip().split("\n")
    # Find the last SBATCH instruction line.
    last_run_sbatch = _last_sbatch_line(run_lines) + 1
    last_setup_sbatch = _last_sbatch_line(setup_lines) + 1
    # Inject environment variables into the script after SBATCH instructions.
    env_lines = [f"export {key}={value}" for key, value in job.envs.items()]
    # Pad the environment variables with newlines.
    env_lines = [""] + env_lines + [""] if env_lines else []
    # Generate the job script.
    # The script should have the following structure:
    # 1. SBATCH instructions from Setup and Run commands (in that order).
    # 2. Environment variables.
    # 3. Setup commands.
    # 4. Run commands.
    output_lines = (
        setup_lines[:last_setup_sbatch]
        + run_lines[:last_run_sbatch]
        + env_lines
        + setup_lines[last_setup_sbatch:]
        + run_lines[last_run_sbatch:]
    )
    # Always start the script with #!/bin/bash.
    script_prefix = "#!/bin/bash"
    if len(output_lines) > 0:
        if not output_lines[0].startswith(script_prefix):
            output_lines.insert(0, script_prefix)
    # Join each line. Always end the script with a new line.
    return "\n".join(output_lines) + "\n"


def _validate_job_config(job: JobConfig) -> None:
    """Validates the provided job configuration.

    Args:
        job: The job to validate.
    """
    if not job.user:
        raise ValueError("User must be provided for Slurm jobs.")
    if not job.run:
        raise ValueError("Run script must be provided for Slurm jobs.")
    if job.num_nodes < 1:
        raise ValueError("Number of nodes must be at least 1.")
    if job.resources.cloud != "slurm":
        raise ValueError(
            f"`Resources.cloud` must be `slurm`. "
            f"Unsupported cloud: {job.resources.cloud}"
        )
    # Warn that other resource parameters are unused for Slurm.
    if not job.working_dir:
        logger.warning("Working directory is not set. This is not recommended.")
    if job.resources.region:
        logger.warning("Region is unused for Slurm jobs.")
    if job.resources.zone:
        logger.warning("Zone is unused for Slurm jobs.")
    if job.resources.accelerators:
        logger.warning("Accelerators are unused for Slurm jobs.")
    if job.resources.cpus:
        logger.warning("CPUs are unused for Slurm jobs.")
    if job.resources.memory:
        logger.warning("Memory is unused for Slurm jobs.")
    if job.resources.instance_type:
        logger.warning("Instance type is unused for Slurm jobs.")
    if job.resources.disk_size:
        logger.warning("Disk size is unused for Slurm jobs.")
    if job.resources.instance_type:
        logger.warning("Instance type is unused for Slurm jobs.")
    # Warn that storage mounts are currently unsupported.
    if len(job.storage_mounts.items()) > 0:
        logger.warning("Storage mounts are currently unsupported for Slurm jobs.")


class SlurmCluster(BaseCluster):
    """A cluster implementation backed by a Slurm scheduler."""

    @dataclass
    class ConnectionInfo:
        """Dataclass to hold information about a connection."""

        hostname: str
        user: str

        @property
        def name(self):
            """Gets the name of the connection in the form user@hostname."""
            return f"{self.user}@{self.hostname}"

    def __init__(self, name: str, client: SlurmClient) -> None:
        """Initializes a new instance of the SlurmCluster class."""
        self._client = client
        self._connection = self.parse_cluster_name(name)

    def __eq__(self, other: Any) -> bool:
        """Checks if two SlurmClusters are equal."""
        if not isinstance(other, SlurmCluster):
            return False
        return self.name() == other.name()

    @staticmethod
    def parse_cluster_name(name: str) -> ConnectionInfo:
        """Parses the cluster name into queue and user components.

        Args:
            name: The name of the cluster.

        Returns:
            _ConnectionInfo: The parsed cluster information.
        """
        # Expected format: <user>@<hostname>
        connection_regex = r"^([a-zA-Z0-9\.\-\_]+)\@([a-zA-Z0-9\.\-\_]+$)"
        match = re.match(connection_regex, name)
        if not match:
            raise ValueError(
                f"Invalid cluster name: {name}. Must be in the format 'user@hostname'."
            )
        return SlurmCluster.ConnectionInfo(hostname=match.group(2), user=match.group(1))

    @staticmethod
    def get_slurm_connections() -> list[ConnectionInfo]:
        """Gets Slurm connections from the OUMI_SLURM_CONNECTIONS env variable."""
        connections_str = os.getenv(_OUMI_SLURM_CONNECTIONS, "")
        if not connections_str:
            return []
        valid_connections = []

        for connection in [h.strip() for h in connections_str.split(",")]:
            try:
                valid_connections.append(SlurmCluster.parse_cluster_name(connection))
            except ValueError:
                logger.warning(
                    f"Invalid Slurm connection string: {connection}. Skipping."
                )
        return valid_connections

    def name(self) -> str:
        """Gets the name of the cluster."""
        return self._connection.name

    def get_job(self, job_id: str) -> Optional[JobStatus]:
        """Gets the jobs on this cluster if it exists, else returns None."""
        for job in self.get_jobs():
            if job.id == job_id:
                return job
        return None

    def get_jobs(self) -> list[JobStatus]:
        """Lists the jobs on this cluster."""
        jobs = self._client.list_jobs()
        for job in jobs:
            job.cluster = self._connection.name
        return jobs

    def cancel_job(self, job_id: str) -> JobStatus:
        """Cancels the specified job on this cluster."""
        self._client.cancel(job_id)
        job = self.get_job(job_id)
        if job is None:
            raise RuntimeError(f"Job {job_id} not found.")
        return job

    def run_job(self, job: JobConfig) -> JobStatus:
        """Runs the specified job on this cluster.

        For Slurm this method consists of 4 parts:

        1. Copy the working directory to ~/oumi_launcher/<submission_time>.
        2. Copy all file mounts.
        3. Create a job script with all env vars, setup, and run commands.
        4. CD into the working directory and submit the job.

        Args:
            job: The job to run.

        Returns:
            JobStatus: The job status.
        """
        _validate_job_config(job)
        job_name = job.name or uuid.uuid1().hex
        submission_time = _format_date(datetime.now())
        remote_working_dir = Path(f"~/oumi_launcher/{submission_time}")
        # Copy the working directory to ~/oumi_launcher/...
        if job.working_dir:
            self._client.put_recursive(job.working_dir, str(remote_working_dir))
        else:
            self._client.run_commands([f"mkdir -p {remote_working_dir}"])
        # Copy all file mounts.
        for remote_path, local_path in job.file_mounts.items():
            self._client.put_recursive(local_path, remote_path)
        # Create the job script by merging envs, setup, and run commands.
        job_script = _create_job_script(job)
        script_path = remote_working_dir / "oumi_job.sh"
        self._client.put(job_script, str(script_path))
        # Set the proper CHMOD permissions.
        self._client.run_commands([f"chmod +x {script_path}"])
        # Submit the job.
        job_id = self._client.submit_job(
            str(script_path),
            str(remote_working_dir),
            job.num_nodes,
            name=job_name,
        )
        max_retries = 3
        wait_time = 5
        for _ in range(max_retries):
            job_status = self.get_job(job_id)
            if job_status is not None:
                return job_status
            logger.info(f"Job {job_id} not found. Retrying in {wait_time} seconds.")
            time.sleep(wait_time)
        job_status = self.get_job(job_id)
        if job_status is None:
            raise RuntimeError(f"Job {job_id} not found after submission.")
        return job_status

    def stop(self) -> None:
        """This is a no-op for Slurm clusters."""
        pass

    def down(self) -> None:
        """This is a no-op for Slurm clusters."""
        pass

    def get_logs_stream(
        self, cluster_name: str, job_id: Optional[str] = None
    ) -> SlurmLogStream:
        """Gets a stream that tails the logs of the target job.

        Args:
            cluster_name: The name of the cluster the job was run in.
            job_id: The ID of the job to tail the logs of.

        Returns:
            A SlurmLogStream object that can be used to read the logs.
        """
        return self._client.get_logs_stream(cluster_name, job_id)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional


class EvaluationResult:
    """Class that retains the evaluation results generated by all backends.

    Attributes:
        task_name: The name of the task on which the model was evaluated.
        task_result: The result of evaluating on the task. This is a dictionary where
            the keys are metric names and the values are their corresponding values.
            The captured metrics vary based on the backend and the specific task.
        backend_config: The configuration of the backend. This is a dictionary including
            configuration parameters that are specific to the backend used to evaluate
            on the task. They are retained for reproducibility.
        start_time: A human-friendly string (recommended format is: YYYYMMDD_HHMMSS)
            that indicates the date and time when the evaluation started.
        elapsed_time_sec: The duration to complete the evaluation (in seconds).
    """

    def __init__(
        self,
        task_name: Optional[str] = None,
        task_result: Optional[dict[str, Any]] = None,
        backend_config: Optional[dict[str, Any]] = None,
        start_time: Optional[str] = None,
        elapsed_time_sec: Optional[int] = None,
    ):
        """Initialize the EvaluationResult class."""
        self.task_name = task_name
        self.task_result = task_result or {}
        self.backend_config = backend_config or {}
        self.start_time = start_time
        self.elapsed_time_sec = elapsed_time_sec

    def to_dict(self) -> dict[str, Any]:
        """Convert the EvaluationResult to a dictionary."""
        return {
            "task_name": self.task_name,
            "task_result": self.task_result,
            "backend_config": self.backend_config,
            "start_time": self.start_time,
            "elapsed_time_sec": self.elapsed_time_sec,
        }

    def get_results(self) -> dict[str, Any]:
        """Retrieves the dict of results."""
        if not self.task_result:
            return {}

        if (
            "results" not in self.task_result
            or self.task_name not in self.task_result["results"]
        ):
            raise ValueError("Unknown `task_result` format.")

        return self.task_result["results"][self.task_name]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oumi.core.datasets import BaseExperimentalKtoDataset
from oumi.core.registry import register_dataset


@register_dataset("trl-lib/kto-mix-14k")
class KtoMix14kDataset(BaseExperimentalKtoDataset):
    """Preprocess the KTO dataset.

    A KTO-formatted version of argilla/dpo-mix-7k designed for Kahneman-Tversky
    Optimization training. This dataset provides binary preference data for
    training language models with human preferences.

    Data Fields:
        - prompt: List of message dictionaries with a single user message
          Example: [{"content": "Question text", "role": "assistant"}]
        - completion: List of message dictionaries with a single assistant message
          Example: [{"content": "Answer text", "role": "assistant"}]
        - label: boolean (True for desirable, False for undesirable)

    See Also:
        For more information on how to use this dataset, refer to:
        - Huggingface hub: https://huggingface.co/datasets/trl-lib/kto-mix-14k
        - KTO documentation: https://huggingface.co/docs/trl/main/en/kto_trainer
    """

    default_dataset = "trl-lib/kto-mix-14k"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import uuid
from copy import deepcopy
from typing import Any, Optional

from oumi.core.configs import JobConfig
from oumi.core.launcher import BaseCluster, JobStatus
from oumi.launcher.clients.local_client import LocalClient
from oumi.utils.logging import logger


def _validate_job_config(job: JobConfig) -> None:
    """Validates the provided job configuration.

    Args:
        job: The job to validate.
    """
    if not job.working_dir:
        raise ValueError("Working directory must be provided for local jobs.")
    if not job.run:
        raise ValueError("Run script must be provided for local jobs.")
    if job.resources.cloud != "local":
        raise ValueError(
            f"`Resources.cloud` must be `local`. "
            f"Unsupported cloud: {job.resources.cloud}"
        )
    # Warn that other resource parameters are unused for local jobs.
    if job.resources.region:
        logger.warning("Region is unused for local jobs.")
    if job.resources.zone:
        logger.warning("Zone is unused for local jobs.")
    if job.resources.accelerators:
        logger.warning("Accelerators are unused for local jobs.")
    if job.resources.cpus:
        logger.warning("CPUs are unused for local jobs.")
    if job.resources.memory:
        logger.warning("Memory is unused for local jobs.")
    if job.resources.instance_type:
        logger.warning("Instance type is unused for local jobs.")
    if job.resources.disk_size:
        logger.warning("Disk size is unused for local jobs.")
    if job.resources.instance_type:
        logger.warning("Instance type is unused for local jobs.")
    # Warn that storage mounts are currently unsupported.
    if len(job.storage_mounts.items()) > 0:
        logger.warning("Storage mounts are currently unsupported for local jobs.")
    if len(job.file_mounts.items()) > 0:
        logger.warning("File mounts are currently unsupported for local jobs.")


class LocalCluster(BaseCluster):
    """A cluster implementation for running jobs locally."""

    def __init__(self, name: str, client: LocalClient) -> None:
        """Initializes a new instance of the LocalCluster class."""
        self._name = name
        self._client = client

    def __eq__(self, other: Any) -> bool:
        """Checks if two LocalClusters are equal."""
        if not isinstance(other, LocalCluster):
            return False
        return self.name() == other.name()

    def name(self) -> str:
        """Gets the name of the cluster."""
        return self._name

    def get_job(self, job_id: str) -> Optional[JobStatus]:
        """Gets the jobs on this cluster if it exists, else returns None."""
        for job in self.get_jobs():
            if job.id == job_id:
                return job
        return None

    def get_jobs(self) -> list[JobStatus]:
        """Lists the jobs on this cluster."""
        jobs = self._client.list_jobs()
        for job in jobs:
            job.cluster = self._name
        return jobs

    def cancel_job(self, job_id: str) -> JobStatus:
        """Cancels the specified job on this cluster."""
        self._client.cancel(job_id)
        job = self.get_job(job_id)
        if job is None:
            raise RuntimeError(f"Job {job_id} not found.")
        return job

    def run_job(self, job: JobConfig) -> JobStatus:
        """Runs the specified job on this cluster.

        Args:
            job: The job to run.

        Returns:
            The job status.
        """
        job_copy = deepcopy(job)
        _validate_job_config(job_copy)
        if not job_copy.name:
            job_copy.name = uuid.uuid1().hex
        status = self._client.submit_job(job_copy)
        status.cluster = self._name
        return status

    def stop(self) -> None:
        """Cancels all jobs, running or queued."""
        for job in self.get_jobs():
            self.cancel_job(job.id)

    def down(self) -> None:
        """Cancels all jobs, running or queued."""
        for job in self.get_jobs():
            self.cancel_job(job.id)

    def get_logs_stream(
        self, cluster_name: str, job_id: Optional[str] = None
    ) -> io.TextIOBase:
        """Gets a stream that tails the logs of the target job.

        Args:
            cluster_name: The name of the cluster the job was run in.
            job_id: The ID of the job to tail the logs of.
        """
        raise NotImplementedError



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from tqdm import tqdm
from typing_extensions import override

from oumi.core.constants import LABEL_IGNORE_INDEX
from oumi.core.datasets.base_map_dataset import BaseMapDataset
from oumi.core.datasets.base_sft_dataset import BaseSftDataset
from oumi.utils.logging import logger


class PackedSftDataset(BaseMapDataset):
    """A dataset that packs samples from a base SFT dataset to maximize efficiency."""

    def __init__(
        self,
        base_dataset: BaseSftDataset,
        max_seq_len: int,
        show_progress: bool = True,
        split_samples: bool = False,
        concat_token_id: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        enable_padding: bool = True,
        **kwargs,
    ):
        """Initialize the PackedSftDataset.

        Args:
            base_dataset: The base SFT dataset to pack samples from.
            max_seq_len: Maximum sequence length for packed samples.
            show_progress: Whether to show progress bar during packing.
                Defaults to True.
            split_samples: Whether to split samples that are longer than max_seq_len.
                If False, samples longer than max_seq_len will be skipped.
                Defaults to False.
            concat_token_id: Token ID to use for concatenating samples.
                If None, samples will be concatenated without a separator token.
                Defaults to None.
            pad_token_id: Token ID to use for padding.
                Required if enable_padding is True. Defaults to None.
            enable_padding: Whether to pad sequences to max_seq_len.
                If True, pad_token_id must be provided. Defaults to True.
            **kwargs: Additional arguments passed to BaseMapDataset.
        """
        super().__init__(**kwargs, dataset_name=base_dataset.dataset_name)

        self.base_dataset = base_dataset

        self._max_seq_len = max_seq_len
        self._disable_tqdm = not show_progress
        self._split_samples = split_samples
        self._concat_token_id = concat_token_id
        self._pad_token_id = pad_token_id
        self._enable_padding = enable_padding
        self._data: list[dict[str, torch.Tensor]] = []

        if self._enable_padding and self._pad_token_id is None:
            raise ValueError(
                "`pad_token_id` must be provided if `enable_padding` is True"
            )

        self._check_dataset_compatibility()
        self._load_data()

    @override
    def _load_data(self) -> None:
        """Pack the base dataset into constant-length samples."""
        buffer = self._get_empty_buffer()

        iterator = range(len(self.base_dataset))

        for idx in tqdm(
            iterator,
            desc="Packing dataset",
            dynamic_ncols=True,
            disable=self._disable_tqdm,
        ):
            sample = self.base_dataset[idx]
            sample_len = len(sample["input_ids"])

            if sample_len > self._max_seq_len and not self._split_samples:
                # We can't split samples, and the sample is too long to fit in
                # the context window. There is no way to handle this sample
                logger.warning(
                    f"Dataset sample is too long ({sample_len} > {self._max_seq_len}). "
                    "Please set `split_samples=True` or increase `max_seq_len`. "
                    "This sample will be skipped."
                )
                continue

            if (
                self._get_potential_sample_len(sample=sample, buffer=buffer)
                == self._max_seq_len
            ):
                # Done with the current buffer, we need to create a new pack
                self._append_sample_to_buffer(sample=sample, buffer=buffer)
                self._append_packed_sample_to_dataset(buffer)
                buffer = self._get_empty_buffer()
                continue
            elif (
                self._get_potential_sample_len(sample=sample, buffer=buffer)
                < self._max_seq_len
            ):
                # We still have space in the buffer, so we add the sample to it
                # and keep going
                self._append_sample_to_buffer(sample=sample, buffer=buffer)
                continue

            # We don't have space in the buffer, so we need to create a new pack
            if self._split_samples:
                self._append_sample_to_buffer(sample=sample, buffer=buffer)

                while self._get_sample_len(buffer) >= self._max_seq_len:
                    finished_sample, buffer = self._split_sample(
                        buffer, cutoff=self._max_seq_len
                    )
                    self._append_packed_sample_to_dataset(finished_sample)
            else:
                # We're not allow to split samples, but buffer + sample is too large
                if self._get_sample_len(buffer) == 0:
                    self._append_sample_to_buffer(sample=sample, buffer=buffer)
                    self._append_packed_sample_to_dataset(buffer)
                else:
                    self._append_packed_sample_to_dataset(buffer)
                    buffer = self._get_empty_buffer()
                    self._append_sample_to_buffer(sample=sample, buffer=buffer)

        # Handle remaining samples in buffer
        if self._get_sample_len(buffer) > 0:
            if self._split_samples:
                while self._get_sample_len(buffer) > 0:
                    finished_sample, buffer = self._split_sample(
                        buffer, cutoff=self._max_seq_len
                    )
                    self._append_packed_sample_to_dataset(finished_sample)
            else:
                self._append_packed_sample_to_dataset(buffer)

    @override
    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        """Get a pack from the dataset by index."""
        if idx >= len(self):
            raise IndexError(f"Index {idx} is out of bounds for PackedSftDataset")
        return self._data[idx]

    @override
    def transform(self, example: dict) -> dict:
        """No-op transform."""
        return example

    #
    # Private methods
    #
    def _append_packed_sample_to_dataset(self, buffer: dict[str, list]) -> None:
        """Creates a fixed length training sample from the buffer and add to dataset."""
        buffer_len = self._get_sample_len(buffer)

        if buffer_len > self._max_seq_len:
            raise ValueError(
                "Buffer is too long "
                f"({buffer_len} >= {self._max_seq_len}). "
                "Please increase `max_seq_len`."
            )

        # Convert lists to tensors
        sample = {k: torch.tensor(v, dtype=torch.long) for k, v in buffer.items()}

        # Pad if needed
        if self._enable_padding and self._pad_token_id is not None:
            if buffer_len < self._max_seq_len:
                pad_length = self._max_seq_len - buffer_len
                for name, value in sample.items():
                    if name == "labels":
                        pad_value = LABEL_IGNORE_INDEX
                    else:
                        pad_value = self._pad_token_id

                    sample[name] = torch.cat(
                        [
                            sample[name],
                            torch.full(
                                (pad_length,), fill_value=pad_value, dtype=torch.long
                            ),
                        ]
                    )

        self._data.append(sample)

    def _append_sample_to_buffer(
        self, sample: dict[str, list], buffer: dict[str, list]
    ) -> None:
        """Append a single training sample to the buffer.

        If concat token is enabled, and if and only if we actually concatenate
        two samples, we add the concat token in between the two samples
        """
        if len(sample["input_ids"]) == 0:
            # Nothing to add
            return

        should_add_concat_token = self._concat_token_id is not None

        if len(buffer["input_ids"]) == 0:
            # Buffer is empty, so we're not concatenating two different samples
            # no need to add concat token
            should_add_concat_token = False

        if should_add_concat_token:
            buffer["input_ids"].append(self._concat_token_id)
            buffer["labels"].append(LABEL_IGNORE_INDEX)  # exclude from loss

        buffer["input_ids"].extend(sample["input_ids"])
        buffer["labels"].extend(sample["labels"])

    def _split_sample(
        self, sample: dict[str, list], cutoff: int
    ) -> tuple[dict[str, list], dict[str, list]]:
        """Split a sample into two parts at the cutoff point.

        Args:
            sample: Dictionary containing lists to split
            cutoff: Index at which to split the lists

        Returns:
            Tuple of two dictionaries containing the split lists
        """
        first_half = {k: v[:cutoff] for k, v in sample.items()}
        second_half = {k: v[cutoff:] for k, v in sample.items()}
        return first_half, second_half

    def _get_empty_buffer(self) -> dict[str, list]:
        """Get an empty buffer with all required fields."""
        return {
            "input_ids": [],
            "labels": [],
        }

    def _get_sample_len(self, buffer: dict[str, list]) -> int:
        """Get the length of the samples in the buffer."""
        return len(buffer["input_ids"])

    def _get_potential_sample_len(
        self, sample: dict[str, list], buffer: dict[str, list]
    ) -> int:
        """Get the length of the samples in the buffer."""
        buffer_len = self._get_sample_len(buffer)
        sample_len = self._get_sample_len(sample)

        # In case we don't need to add a concat token
        if self._concat_token_id is None or buffer_len == 0 or sample_len == 0:
            return buffer_len + sample_len

        # In case we need to add a concat token
        return buffer_len + sample_len + 1

    def _check_dataset_compatibility(self) -> None:
        """Check the base dataset for errors."""
        if len(self.base_dataset) == 0:
            raise ValueError("Base dataset is empty. Cannot pack empty dataset.")

        keys = set(self.base_dataset[0].keys())

        if "input_ids" not in keys:
            raise ValueError("Base dataset must contain 'input_ids' key.")

        if "labels" not in keys:
            raise ValueError("Base dataset must contain 'labels' key.")

        if set(keys) != {"input_ids", "labels"}:
            logger.warning(
                "Base dataset contains additional keys. "
                "Only 'input_ids' and 'labels' will be used."
            )



from typing import Optional
from unittest.mock import Mock, patch

import pytest
from torch import nn

from oumi.builders.models import (
    _get_model_type,
    _patch_model_for_liger_kernel,
    build_chat_template,
    build_huggingface_model,
    build_tokenizer,
    is_image_text_llm,
)
from oumi.core.configs import ModelParams
from oumi.core.configs.internal.supported_models import find_model_hf_config
from oumi.utils.logging import logger


@pytest.fixture
def mock_liger_kernel():
    with patch("oumi.builders.models.liger_kernel") as mock:
        yield mock


def create_mock_model(model_type):
    model = Mock(spec=nn.Module)
    model.config = Mock()
    model.config.model_type = model_type
    return model


@pytest.mark.parametrize(
    "model_type",
    [
        "llama",
        "mixtral",
        "mistral",
        "gemma",
    ],
)
def test_patch_model_for_liger_kernel(mock_liger_kernel, model_type):
    model = create_mock_model(model_type)

    _patch_model_for_liger_kernel(model)

    mock_liger_kernel.transformers._apply_liger_kernel.assert_called_once_with(
        model_type
    )


def test_patch_model_for_liger_kernel_no_config(mock_liger_kernel):
    model = Mock(spec=nn.Module)
    with pytest.raises(ValueError, match=f"Could not find model type for: {model}"):
        _patch_model_for_liger_kernel(model)


def test_patch_model_for_liger_kernel_import_error():
    with patch("oumi.builders.models.liger_kernel", None):
        model = create_mock_model("llama")
        with pytest.raises(ImportError, match="Liger Kernel not installed"):
            _patch_model_for_liger_kernel(model)


def test_get_model_type():
    # Test with valid model
    model = create_mock_model("llama")
    assert _get_model_type(model) == "llama"

    # Test with no config
    model = Mock(spec=nn.Module)
    assert _get_model_type(model) is None

    # Test with config but no model_type
    model = Mock(spec=nn.Module)
    model.config = Mock()
    model.config.model_type = None
    assert _get_model_type(model) is None


@pytest.mark.parametrize(
    "template_name",
    ["zephyr"],
)
def test_build_chat_template_existing_templates(template_name):
    template = build_chat_template(template_name)

    assert template is not None
    assert len(template) > 0


def test_build_chat_template_nonexistent_template():
    with pytest.raises(FileNotFoundError) as exc_info:
        build_chat_template("nonexistent_template")

    assert "Chat template file not found" in str(exc_info.value)


def test_build_chat_template_removes_indentation_and_newlines():
    template_content = """
        {{ bos_token }}
        {% for message in messages %}
            {% if message['role'] == 'user' %}
                User: {{ message['content'] }}
            {% elif message['role'] == 'assistant' %}
                Assistant: {{ message['content'] }}
            {% endif %}
            {{ eos_token }}
        {% endfor %}
    """
    expected = (
        "{{ bos_token }}{% for message in messages %}{% if message['role'] == 'user' %}"
        "User: {{ message['content'] }}{% elif message['role'] == 'assistant' %}"
        "Assistant: {{ message['content'] }}{% endif %}{{ eos_token }}{% endfor %}"
    )

    with (
        patch("oumi.builders.models.get_oumi_root_directory"),
        patch("oumi.builders.models.load_file") as mock_load_file,
    ):
        mock_load_file.return_value = template_content

        result = build_chat_template("test_template")

        assert result == expected
        mock_load_file.assert_called_once()


@pytest.mark.parametrize(
    "model_name, trust_remote_code, expected_result",
    [
        ("MlpEncoder", False, False),  # Custom model
        ("CnnClassifier", False, False),  # Custom model
        ("openai-community/gpt2", False, False),
        ("HuggingFaceTB/SmolLM2-135M-Instruct", False, False),
        ("llava-hf/llava-1.5-7b-hf", False, True),
        ("Salesforce/blip2-opt-2.7b", False, True),
        ("microsoft/Phi-3-vision-128k-instruct", True, True),
        ("HuggingFaceTB/SmolVLM-Instruct", False, True),
    ],
)
def test_is_image_text_llm(
    model_name: str, trust_remote_code: bool, expected_result: bool
):
    assert (
        is_image_text_llm(
            ModelParams(model_name=model_name, trust_remote_code=trust_remote_code)
        )
        == expected_result
    )


@pytest.mark.parametrize(
    "model_name, trust_remote_code, template_name, expected_padding_side",
    [
        ("openai-community/gpt2", False, "gpt2", "right"),
        ("HuggingFaceTB/SmolLM2-135M-Instruct", False, None, "right"),
        ("llava-hf/llava-1.5-7b-hf", False, "llava", "left"),
        ("microsoft/Phi-3-vision-128k-instruct", True, "phi3-instruct", "right"),
        ("Qwen/Qwen2-VL-2B-Instruct", True, "qwen2-vl-instruct", "left"),
        # These models require allowlisting:
        # ("meta-llama/Llama-3.2-3B-Instruct", False, None, "right"),
    ],
)
def test_default_chat_template_in_build_tokenizer(
    model_name: str,
    trust_remote_code: bool,
    template_name: Optional[str],
    expected_padding_side: str,
):
    tokenizer = build_tokenizer(
        ModelParams(model_name=model_name, trust_remote_code=trust_remote_code)
    )

    debug_tag = f"template_name: {template_name} model_name: {model_name}"
    if template_name:
        expected = build_chat_template(template_name=template_name)
        if tokenizer.chat_template != expected:
            logger.info(
                f"Tokenizer chat template:\n\n{tokenizer.chat_template}\n\n"
                f"Expected chat template:\n\n{expected}\n\n"
            )
        assert tokenizer.chat_template == expected, debug_tag
    else:
        # Using the model's built-in config.
        assert tokenizer.chat_template is not None, (
            f"Unspecified built-in template: {debug_tag}"
        )
        assert len(tokenizer.chat_template) > 0, f"Empty built-in template: {debug_tag}"

    # Also check padding side here.
    assert hasattr(tokenizer, "padding_side")
    assert tokenizer.padding_side == expected_padding_side


def test_find_model_hf_config_with_custom_kwargs():
    mock_config = Mock()
    mock_config.model_type = "test_model"

    custom_kwargs = {
        "max_position_embeddings": 2048,
    }

    with patch(
        "oumi.core.configs.internal.supported_models.transformers.AutoConfig.from_pretrained"
    ) as mock_from_pretrained:
        mock_from_pretrained.return_value = (mock_config, {})

        result = find_model_hf_config(
            model_name="test-model",
            trust_remote_code=False,
            revision="main",
            **custom_kwargs,
        )

        # Verify the result
        assert result == mock_config

        # Verify AutoConfig.from_pretrained was called with custom kwargs
        mock_from_pretrained.assert_called_once_with(
            "test-model",
            trust_remote_code=False,
            return_unused_kwargs=True,
            revision="main",
            **custom_kwargs,
        )


def test_find_model_hf_config_logs_unused_kwargs():
    """Test that find_model_hf_config logs a warning for unused kwargs."""
    mock_config = Mock()
    mock_config.model_type = "test_model"
    unused_kwargs = {"unsupported_param": "value"}

    with (
        patch(
            "oumi.core.configs.internal.supported_models.transformers.AutoConfig.from_pretrained"
        ) as mock_from_pretrained,
        patch("oumi.core.configs.internal.supported_models.logger") as mock_logger,
    ):
        mock_from_pretrained.return_value = (mock_config, unused_kwargs)

        find_model_hf_config(
            model_name="test-model",
            trust_remote_code=False,
            unsupported_param="value",
        )

        # Verify warning was logged
        mock_logger.warning.assert_called_once_with(
            f"Unused kwargs found in 'test-model' config: {unused_kwargs}."
        )


def test_build_huggingface_model_passes_model_kwargs_to_find_model_hf_config():
    """Test that build_huggingface_model passes model_kwargs to find_model_hf_config."""
    model_kwargs = {
        "max_position_embeddings": 2048,
    }

    model_params = ModelParams(
        model_name="test-model", trust_remote_code=False, model_kwargs=model_kwargs
    )

    mock_config = Mock()
    mock_config.model_type = "llama"
    mock_config.use_cache = True

    mock_model = Mock()
    mock_model.config = mock_config

    with (
        patch("oumi.builders.models.find_model_hf_config") as mock_find_config,
        patch("oumi.builders.models._get_transformers_model_class") as mock_get_class,
        patch("oumi.builders.models.get_device_rank_info") as mock_device_info,
        patch("oumi.builders.models.is_using_accelerate_fsdp", return_value=False),
    ):
        mock_find_config.return_value = mock_config
        mock_model_class = Mock()
        mock_model_class.from_pretrained.return_value = mock_model
        mock_get_class.return_value = mock_model_class
        mock_device_info.return_value = Mock(world_size=1, local_rank=0)

        result = build_huggingface_model(model_params)

        # Verify find_model_hf_config was called with model_kwargs
        mock_find_config.assert_called_once_with(
            "test-model", trust_remote_code=False, revision=None, **model_kwargs
        )

        # Verify the model was built successfully
        assert result == mock_model



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import torch
from typing_extensions import override

from oumi.core.datasets.base_iterable_dataset import BaseIterableDataset
from oumi.core.tokenizers import BaseTokenizer


class BasePretrainingDataset(BaseIterableDataset):
    """Base class for pretraining iterable datasets.

    This class extends BaseIterableDataset to provide functionality specific to
    pretraining tasks.

    Attributes:
        tokenizer (BaseTokenizer): The tokenizer used for text encoding.
        seq_length (int): The desired sequence length for model inputs.
        concat_token_id (int): The ID of the token used to concatenate documents.

    Example:
        >>> from transformers import AutoTokenizer
        >>> from oumi.builders import build_tokenizer
        >>> from oumi.core.configs import ModelParams
        >>> from oumi.core.datasets import BasePretrainingDataset
        >>> tokenizer = build_tokenizer(ModelParams(model_name="gpt2"))
        >>> dataset = BasePretrainingDataset(
        ...     dataset_name="wikimedia/wikipedia",
        ...     subset="20231101.en",
        ...     split="train",
        ...     tokenizer=tokenizer,
        ...     seq_length=512
        ... )
        >>> example = next(iter(dataset))
    """

    def __init__(
        self,
        *,
        tokenizer: BaseTokenizer,
        seq_length: int,
        dataset_text_field: str = "text",
        append_concat_token: bool = True,
        add_special_tokens: bool = True,
        skip_last: bool = True,
        **kwargs,
    ):
        """Initializes a new instance of the BasePretrainingDataset class."""
        if append_concat_token and tokenizer.eos_token_id is None:
            raise ValueError(
                "Tokenizer must have an EOS token if append_concat_token is enabled."
            )

        self.concat_token_id = tokenizer.eos_token_id if append_concat_token else None

        self.tokenizer = tokenizer
        self.seq_length = seq_length
        self._dataset_text_field = dataset_text_field
        self._append_concat_token = append_concat_token
        self._add_special_tokens = add_special_tokens
        self._skip_last = skip_last

        super().__init__(**kwargs)

    def __iter__(self):
        """Iterates over the dataset and yields samples of a specified sequence length.

        The underlying dataset is a stream of documents. Each document is expected to
        contain a text field `self._dataset_text_field` that will be tokenized.
        Training samples are then yielded in sequences of length `self.seq_length`.

        Given this iterator might return samples from different documents, we optionally
        use `self.concat_token_id` to separate the sequences from different documents.
        """
        buffer = []
        for document in self.data:
            if self._append_concat_token and len(buffer) > 0:
                # We started preprocessing a new document
                # so we need to append the concatenation token to mark the end
                # of the previous document.
                buffer.append(self.concat_token_id)

            # Pre-process and tokenize the document
            document_tokens = self.transform(document[self._dataset_text_field])
            buffer.extend(document_tokens)

            # Yield sequences of the specified length.
            # Otherwise, resume pre-processing the next document.
            while len(buffer) >= self.seq_length:
                # We have enough tokens to create a fully packed sample
                yield self._create_training_sample(buffer[: self.seq_length])
                buffer = buffer[self.seq_length :]

        # Finished iterating on the dataset, yield the remaining buffer
        if len(buffer) > 0:
            if not self._skip_last or len(buffer) == self.seq_length:
                yield self._create_training_sample(buffer)

    @override
    def transform(self, sample: Any) -> list[int]:
        """Preprocesses the inputs in the given sample."""
        return self.tokenize(sample)

    def tokenize(self, text: str) -> list[int]:
        """Tokenizes the given text.

        Should not apply any padding/truncation to allow for packing.
        """
        return self.tokenizer.encode(
            text=text,
            return_tensors=None,
            max_length=None,
            padding=False,
            truncation=False,
            add_special_tokens=self._add_special_tokens,
        )

    def _create_training_sample(self, tokens: list) -> dict[str, torch.Tensor]:
        """Creates a training sample from the given tokens."""
        input_ids = torch.tensor(tokens)
        attention_mask = torch.ones_like(input_ids)
        return {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "labels": input_ids.clone(),
        }
